-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author: E. Kooistra
-- Purpose: Immediately apply reset and synchronously release it at rising clk
-- Description:
--   When in_rst gets asserted, then the out_rst gets asserted immediately (= asynchronous reset apply).
--   When in_rst gets de-assered, then out_rst gets de-asserted after g_delay_len cycles (= synchronous reset release).
--
--   The in_rst assert level is set by g_in_rst_level.
--   The out_rst assert level is set by c_out_rst_level = g_rst_level.
--
-- Remarks:
-- . The in_rst can also synchronise other signals than a reset, e.g. a locked signal from a PLL.

library IEEE;
use IEEE.std_logic_1164.all;
use work.common_pkg.all;

entity common_areset is
  generic (
    g_in_rst_level : std_logic := '1';  -- = in_rst level
    g_rst_level    : std_logic := '1';  -- = out_rst level (keep original generic
                                        --   name for backward compatibility)
    g_delay_len    : natural   := c_meta_delay_len
  );
  port (
    in_rst    : in  std_logic;
    clk       : in  std_logic;
    out_rst   : out std_logic
  );
end;

architecture str of common_areset is
  constant c_out_rst_level   : std_logic := g_rst_level;
  constant c_out_rst_level_n : std_logic := not g_rst_level;

  signal i_rst               : std_logic;
begin
  i_rst <= in_rst when g_in_rst_level = '1' else not in_rst;

  u_async : entity work.common_async
  generic map (
    g_rst_level => c_out_rst_level,
    g_delay_len => g_delay_len
  )
  port map (
    rst  => i_rst,
    clk  => clk,
    din  => c_out_rst_level_n,
    dout => out_rst
  );
end str;
