-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: MM data buffer and Rx seq for multiple parallel SOSI streams
-- Description:
-- . g_use_db
--   The mms_diag_data_buffer can capture data from an input stream in a data
--   buffer when g_use_db=TRUE. Dependend on g_buf_use_sync the data buffer
--   is rewritten after each in_sync or when the last word was read via MM.
-- . g_use_rx_seq
--   The mms_diag_data_buffer can continously verify a input Rx data sequence
--   when g_use_rx_seq=TRUE. The expected sequence data is typically generated
--   by an remote upstream tx_seq source.
-- . The advantage of the rx_seq is that is can continously verify the
--   correctness of all rx data in hardware, whereas the DB can only take a
--   snapshot that then needs to be examined via MM. The advandage of the DB
--   is that it can take a snapshot of the values of the received data. The
--   DB requires RAM resources and the rx_seq does not.
--
-- Block diagram:
--
--                           g_use_db
--                           g_buf_use_sync
--                              .
--                              .      g_use_tx_seq
--                              .          .
--                              .          .
--                      /-------------> Rx seq
--                      |       .         |
--     in_sosi_arr -----*---> DB RAM      |
--     in_sync -------------> DB reg      |
--                              ||        |
--                              ||        |
--              MM ================================
--
-- Remark:
-- . A nice new feature would be to continuously write the DB and to stop
--   writting it on a trigger. This trigger can then eg. be when the rx_seq
--   detects an error. By delaying the trigger somewhat it the DB can then
--   capture some data before and after the trigger event.

library IEEE, common_lib, technology_lib, dp_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_mem_pkg.all;
use dp_lib.dp_stream_pkg.all;
use work.diag_pkg.all;
use technology_lib.technology_select_pkg.all;

entity mms_diag_data_buffer_dev is
  generic (
    g_technology   : natural := c_tech_select_default;
    -- Generate configurationsc_diag_db
    g_use_db       : boolean := true;
    g_use_rx_seq   : boolean := false;
    -- General
    g_nof_streams  : positive := 16;  -- each stream gets an data buffer
    -- DB settings
    g_data_type    : t_diag_data_type_enum := e_data;  -- define the sosi field that gets stored: e_data=data, e_complex=im&re, e_real=re, e_imag=im
    g_data_w       : natural := 32;  -- the g_data_w is the width of the data, re, im values or of the combined im&re value
    g_buf_nof_data : natural := 1024;  -- nof words per data buffer
    g_buf_use_sync : boolean := false;  -- when TRUE start filling the buffer at the in_sync, else after the last word was read
    -- Rx_seq
    g_use_steps    : boolean := false;
    g_nof_steps    : natural := c_diag_seq_rx_reg_nof_steps;
    g_seq_dat_w    : natural := 32  -- >= 1, test sequence data width. Choose g_seq_dat_w <= g_data_w
  );
  port (
    -- System
    mm_rst            : in  std_logic;
    mm_clk            : in  std_logic;
    dp_rst            : in  std_logic;
    dp_clk            : in  std_logic;
    -- MM interface
    reg_data_buf_mosi : in  t_mem_mosi := c_mem_mosi_rst;  -- DB control register (one per stream)
    reg_data_buf_miso : out t_mem_miso;

    ram_data_buf_mosi : in  t_mem_mosi := c_mem_mosi_rst;  -- DB buffer RAM (one per streams)
    ram_data_buf_miso : out t_mem_miso;

    reg_rx_seq_mosi   : in  t_mem_mosi := c_mem_mosi_rst;  -- Rx seq control register (one per streams)
    reg_rx_seq_miso   : out t_mem_miso;

    -- ST interface
    in_sync           : in  std_logic := '0';  -- input sync pulse in ST dp_clk domain that starts data buffer when g_use_in_sync = TRUE
    in_sosi_arr       : in t_dp_sosi_arr(g_nof_streams - 1 downto 0);
    out_wr_done_arr   : out std_logic_vector(g_nof_streams - 1 downto 0)
  );
end mms_diag_data_buffer_dev;

architecture str of mms_diag_data_buffer_dev is
  constant c_buf_mm_factor   : natural := ceil_div(g_data_w, c_word_w);
  constant c_buf_nof_data_mm : natural := g_buf_nof_data * c_buf_mm_factor;

  constant c_buf_adr_w : natural := ceil_log2(c_buf_nof_data_mm);
  constant c_reg_adr_w : natural := c_diag_db_dev_reg_adr_w;

  type t_data_arr is array (integer range <>) of std_logic_vector(g_data_w - 1 downto 0);

  signal in_data_arr           : t_data_arr(g_nof_streams - 1 downto 0);

  signal ram_data_buf_mosi_arr : t_mem_mosi_arr(g_nof_streams - 1 downto 0);
  signal ram_data_buf_miso_arr : t_mem_miso_arr(g_nof_streams - 1 downto 0);

  signal reg_data_buf_mosi_arr : t_mem_mosi_arr(g_nof_streams - 1 downto 0);
  signal reg_data_buf_miso_arr : t_mem_miso_arr(g_nof_streams - 1 downto 0);
begin
  no_db : if g_use_db = false generate
    ram_data_buf_miso <= c_mem_miso_rst;
    reg_data_buf_miso <= c_mem_miso_rst;
  end generate;

  gen_db : if g_use_db = true generate
    -- Combine the internal array of mm interfaces for the data_buf to one array that is connected to the port of the MM bus
    u_mem_mux_data_buf : entity common_lib.common_mem_mux
    generic map (
      g_nof_mosi    => g_nof_streams,
      g_mult_addr_w => c_buf_adr_w
    )
    port map (
      mosi     => ram_data_buf_mosi,
      miso     => ram_data_buf_miso,
      mosi_arr => ram_data_buf_mosi_arr,
      miso_arr => ram_data_buf_miso_arr
    );

    u_mem_mux_reg : entity common_lib.common_mem_mux
    generic map (
      g_nof_mosi    => g_nof_streams,
      g_mult_addr_w => c_reg_adr_w
    )
    port map (
      mosi     => reg_data_buf_mosi,
      miso     => reg_data_buf_miso,
      mosi_arr => reg_data_buf_mosi_arr,
      miso_arr => reg_data_buf_miso_arr
    );

    gen_stream : for I in 0 to g_nof_streams - 1 generate
      in_data_arr(I) <= in_sosi_arr(I).im(g_data_w / 2 - 1 downto 0) & in_sosi_arr(I).re(g_data_w / 2 - 1 downto 0) when g_data_type = e_complex else
                        in_sosi_arr(I).re(g_data_w - 1 downto 0)                                              when g_data_type = e_real else
                        in_sosi_arr(I).im(g_data_w - 1 downto 0)                                              when g_data_type = e_imag else
                        in_sosi_arr(I).data(g_data_w - 1 downto 0);  -- g_data_type=e_data is default

      u_diag_data_buffer : entity work.diag_data_buffer_dev
      generic map (
        g_technology  => g_technology,
        g_data_w      => g_data_w,
        g_nof_data    => g_buf_nof_data,
        g_use_in_sync => g_buf_use_sync  -- when TRUE start filling the buffer at the in_sync, else after the last word was read
      )
      port map (
        -- Memory-mapped clock domain
        mm_rst      => mm_rst,
        mm_clk      => mm_clk,

        ram_mm_mosi => ram_data_buf_mosi_arr(I),
        ram_mm_miso => ram_data_buf_miso_arr(I),

        reg_mm_mosi => reg_data_buf_mosi_arr(I),
        reg_mm_miso => reg_data_buf_miso_arr(I),

        -- Streaming clock domain
        st_rst      => dp_rst,
        st_clk      => dp_clk,

        in_data     => in_data_arr(I),
        in_sync     => in_sync,
        in_val      => in_sosi_arr(I).valid,
        out_wr_done => out_wr_done_arr(I)
      );
    end generate;
  end generate;

  no_rx_seq : if g_use_rx_seq = false generate
    reg_rx_seq_miso <= c_mem_miso_rst;
  end generate;

  gen_rx_seq : if g_use_rx_seq = true generate
    u_mms_diag_rx_seq : entity work.mms_diag_rx_seq
    generic map (
      g_nof_streams => g_nof_streams,
      g_use_steps   => g_use_steps,
      g_nof_steps   => g_nof_steps,
      g_seq_dat_w   => g_seq_dat_w,  -- >= 1, test sequence data width
      g_data_w      => g_data_w  -- >= g_seq_dat_w, user data width
    )
    port map (
      -- Clocks and reset
      mm_rst         => mm_rst,
      mm_clk         => mm_clk,
      dp_rst         => dp_rst,
      dp_clk         => dp_clk,

      -- Memory Mapped Slave
      reg_mosi       => reg_rx_seq_mosi,  -- multiplexed port for g_nof_streams MM control/status registers
      reg_miso       => reg_rx_seq_miso,

      -- Streaming interface
      rx_snk_in_arr  => in_sosi_arr
    );
  end generate;

end str;
