-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . Test dp_offload_tx and dp_offload_rx components
-- Description:
-- . Block generators generate data blocks that flow from dp_offload_tx to dp_offload_rx
--   instances via 1GbE (32b user interface)

library IEEE, common_lib, unb1_board_lib, dp_lib, eth_lib, tech_tse_lib, diag_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_str_pkg.all;
use common_lib.common_mem_pkg.all;
use unb1_board_lib.unb1_board_pkg.all;
use unb1_board_lib.unb1_board_peripherals_pkg.all;
use dp_lib.dp_stream_pkg.all;
use tech_tse_lib.tech_tse_pkg.all;
use eth_lib.eth_pkg.all;
use common_lib.common_network_layers_pkg.all;
use diag_lib.diag_pkg.all;
use common_lib.common_field_pkg.all;

entity unb1_dp_offload is
  generic (
    g_sim         : boolean := false;  -- set by     ModelSim
    g_sim_unb_nr  : natural := 0;  -- set by     ModelSim
    g_sim_node_nr : natural := 0;  -- set by     ModelSim
    g_stamp_date  : natural := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time  : natural := 0;  -- Time (HHMMSS)   -- set by QSF
    g_stamp_svn   : natural := 0  -- SVN revision    -- set by QSF
  );
  port (
    -- GENERAL
--  CLK           : IN    STD_LOGIC; -- dp_clk is generated by SOPC altpll
    PPS           : in    std_logic;
    WDI           : out   std_logic;
    INTA          : inout std_logic;
    INTB          : inout std_logic;

    -- Others
    VERSION       : in    std_logic_vector(c_unb1_board_aux.version_w - 1 downto 0);
    ID            : in    std_logic_vector(c_unb1_board_aux.id_w - 1 downto 0);
    TESTIO        : inout std_logic_vector(c_unb1_board_aux.testio_w - 1 downto 0);

    -- I2C Interface to Sensors
    sens_sc       : inout std_logic;
    sens_sd       : inout std_logic;

    -- 1GbE Control Interface
    ETH_clk       : in    std_logic;
    ETH_SGIN      : in    std_logic;
    ETH_SGOUT     : out   std_logic
  );
end unb1_dp_offload;

architecture str of unb1_dp_offload is
  constant c_design_name                : string  := "unb1_dp_offload";
  constant c_nof_streams                : natural := 1;
  constant c_data_w                     : natural := c_tech_tse_data_w;

  -- Block generator
  constant c_bg_block_size              : natural := 900;
  constant c_bg_gapsize                 : natural := 100;
  constant c_bg_blocks_per_sync         : natural := sel_a_b(g_sim, 10, 200000);  -- 200000*(900+100) = 200000000 cycles = 1 second
  constant c_bg_ctrl                    : t_diag_block_gen := ('0',  -- enable
                                                               '0',  -- enable_sync
                                                              TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                              TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                              TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                              TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                              TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                              TO_UVEC(                   0, c_diag_bg_bsn_init_w));

  -- dp_offload_tx
  constant c_nof_hdr_fields : natural := 4 + 12 + 4 + 9;  -- Total header bits = 512
  constant c_hdr_field_arr  : t_common_field_arr(c_nof_hdr_fields - 1 downto 0) := ( ( field_name_pad("eth_word_align"     ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("eth_dst_mac"        ), "  ", 48, field_default(x"002286080000") ),
                                                                                   ( field_name_pad("eth_src_mac"        ), "  ", 48, field_default(0) ),
                                                                                   ( field_name_pad("eth_type"           ), "  ", 16, field_default(x"0800") ),
                                                                                   ( field_name_pad("ip_version"         ), "  ",  4, field_default(4) ),
                                                                                   ( field_name_pad("ip_header_length"   ), "  ",  4, field_default(5) ),
                                                                                   ( field_name_pad("ip_services"        ), "  ",  8, field_default(0) ),
                                                                                   ( field_name_pad("ip_total_length"    ), "  ", 16, field_default(128) ),
                                                                                   ( field_name_pad("ip_identification"  ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("ip_flags"           ), "  ",  3, field_default(2) ),
                                                                                   ( field_name_pad("ip_fragment_offset" ), "  ", 13, field_default(0) ),
                                                                                   ( field_name_pad("ip_time_to_live"    ), "  ",  8, field_default(127) ),
                                                                                   ( field_name_pad("ip_protocol"        ), "  ",  8, field_default(17) ),
                                                                                   ( field_name_pad("ip_header_checksum" ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("ip_src_addr"        ), "  ", 32, field_default(0) ),
                                                                                   ( field_name_pad("ip_dst_addr"        ), "  ", 32, field_default(0) ),
                                                                                   ( field_name_pad("udp_src_port"       ), "  ", 16, field_default(4000) ),
                                                                                   ( field_name_pad("udp_dst_port"       ), "  ", 16, field_default(4000) ),
                                                                                   ( field_name_pad("udp_total_length"   ), "  ", 16, field_default(108) ),
                                                                                   ( field_name_pad("udp_checksum"       ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("usr_sync"           ), "  ",  1, field_default(0) ),
                                                                                   ( field_name_pad("usr_bsn"            ), "  ", 60, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_0"    ), "  ",  7, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_1"    ), "  ",  9, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_2"    ), "  ", 10, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_3"    ), "  ", 33, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_4"    ), "  ",  5, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_5"    ), "  ",  8, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_6"    ), "  ", 27, field_default(0) ) );

  constant c_hdr_field_ovr_init         : std_logic_vector(c_nof_hdr_fields - 1 downto 0) := "1101" & "111111111100" & "1111" & "001111111";

  constant c_nof_words_per_block        : natural := 11;
  constant c_nof_blocks_per_packet      : natural := 2;

  signal hdr_fields_in_arr              : t_slv_1024_arr(c_nof_streams - 1 downto 0);
  signal hdr_fields_out_arr             : t_slv_1024_arr(c_nof_streams - 1 downto 0);

  -- System
  signal cs_sim                         : std_logic;
  signal xo_clk                         : std_logic;
  signal xo_rst                         : std_logic;
  signal xo_rst_n                       : std_logic;
  signal mm_clk                         : std_logic;
  signal mm_locked                      : std_logic;
  signal mm_rst                         : std_logic;
  signal dp_rst                         : std_logic;
  signal dp_clk                         : std_logic;
  -- PIOs
  signal pout_wdi                       : std_logic;

  signal eth1g_tse_clk                  : std_logic;
  signal eth1g_mm_rst                   : std_logic;
  signal eth1g_tse_mosi                 : t_mem_mosi;
  signal eth1g_tse_miso                 : t_mem_miso;
  signal eth1g_reg_mosi                 : t_mem_mosi;
  signal eth1g_reg_miso                 : t_mem_miso;
  signal eth1g_reg_interrupt            : std_logic;
  signal eth1g_ram_mosi                 : t_mem_mosi;
  signal eth1g_ram_miso                 : t_mem_miso;

  signal reg_wdi_mosi                   : t_mem_mosi;
  signal reg_wdi_miso                   : t_mem_miso;
  signal reg_unb_system_info_mosi       : t_mem_mosi;
  signal reg_unb_system_info_miso       : t_mem_miso;
  signal rom_unb_system_info_mosi       : t_mem_mosi;
  signal rom_unb_system_info_miso       : t_mem_miso;
  signal reg_unb_sens_mosi              : t_mem_mosi;
  signal reg_unb_sens_miso              : t_mem_miso;

  signal reg_diag_bg_mosi               : t_mem_mosi;
  signal reg_diag_bg_miso               : t_mem_miso;
  signal ram_diag_bg_mosi               : t_mem_mosi;
  signal ram_diag_bg_miso               : t_mem_miso;

  signal reg_dp_offload_tx_hdr_dat_mosi : t_mem_mosi;
  signal reg_dp_offload_tx_hdr_dat_miso : t_mem_miso;
  signal reg_dp_offload_rx_hdr_dat_mosi : t_mem_mosi;
  signal reg_dp_offload_rx_hdr_dat_miso : t_mem_miso;

  signal reg_bsn_monitor_mosi           : t_mem_mosi;
  signal reg_bsn_monitor_miso           : t_mem_miso;
  signal ram_diag_data_buf_mosi         : t_mem_mosi;
  signal ram_diag_data_buf_miso         : t_mem_miso;
  signal reg_diag_data_buf_mosi         : t_mem_mosi;
  signal reg_diag_data_buf_miso         : t_mem_miso;

  signal block_gen_src_out_arr          : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal block_gen_src_in_arr           : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  signal dp_offload_tx_snk_in_arr       : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_tx_snk_out_arr      : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_tx_src_out_arr      : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_tx_src_in_arr       : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_rx_snk_in_arr       : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_rx_snk_out_arr      : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_rx_src_out_arr      : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_rx_src_in_arr       : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  signal diag_data_buf_snk_in_arr       : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal diag_data_buf_snk_out_arr      : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_fifo_fill_src_out_arr       : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_fifo_fill_src_in_arr        : t_dp_siso_arr(c_nof_streams - 1 downto 0);
begin
  -----------------------------------------------------------------------------
  -- TX: Block generator
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : entity diag_lib.mms_diag_block_gen
  generic map (
    g_nof_streams        => c_nof_streams,
    g_buf_dat_w          => c_data_w,
    g_buf_addr_w         => ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
    g_file_name_prefix   => "hex/counter_data_" & natural'image(c_data_w),
    g_diag_block_gen_rst => c_bg_ctrl
  )
  port map (
    mm_rst           => mm_rst,
    mm_clk           => mm_clk,

    dp_rst           => dp_rst,
    dp_clk           => dp_clk,

    out_sosi_arr     => block_gen_src_out_arr,
    out_siso_arr     => block_gen_src_in_arr,

    reg_bg_ctrl_mosi => reg_diag_bg_mosi,
    reg_bg_ctrl_miso => reg_diag_bg_miso,
    ram_bg_data_mosi => ram_diag_bg_mosi,
    ram_bg_data_miso => ram_diag_bg_miso
  );

  ---------------------------------------------------------------------------------------
  -- Use a FIFO when the source has no flow control
  -- . unless source is 100% valid, read note in dp_offload_tx
  ---------------------------------------------------------------------------------------
  gen_dp_fifo_sc : for i in 0 to c_nof_streams - 1 generate
    u_dp_fifo_sc : entity dp_lib.dp_fifo_sc
    generic map (
      g_data_w    => c_data_w,
      g_use_bsn   => true,
      g_bsn_w     => 64,
      g_use_sync  => true,
      g_fifo_size => 100
     )
    port map (
      rst        => dp_rst,
      clk        => dp_clk,

      snk_in     => block_gen_src_out_arr(i),
      snk_out    => block_gen_src_in_arr(i),

      src_out    => dp_offload_tx_snk_in_arr(i),
      src_in     => dp_offload_tx_snk_out_arr(i)
    );
  end generate;

  -----------------------------------------------------------------------------
  -- TX: dp_offload_tx
  -----------------------------------------------------------------------------
  u_dp_offload_tx : entity dp_lib.dp_offload_tx
  generic map (
    g_nof_streams               => c_nof_streams,
    g_data_w                    => c_data_w,
    g_use_complex               => false,
    g_nof_words_per_block       => c_nof_words_per_block,
    g_nof_blocks_per_packet     => c_nof_blocks_per_packet,
    g_hdr_field_arr             => c_hdr_field_arr,
    g_hdr_field_sel             => c_hdr_field_ovr_init
   )
  port map (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,

    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

    reg_hdr_dat_mosi      => reg_dp_offload_tx_hdr_dat_mosi,
    reg_hdr_dat_miso      => reg_dp_offload_tx_hdr_dat_miso,

    snk_in_arr            => dp_offload_tx_snk_in_arr,
    snk_out_arr           => dp_offload_tx_snk_out_arr,

    src_out_arr           => dp_offload_tx_src_out_arr,
    src_in_arr            => dp_offload_tx_src_in_arr,

    hdr_fields_in_arr     => hdr_fields_in_arr
  );

  gen_hdr_in_fields : for i in 0 to c_nof_streams - 1 generate
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "eth_src_mac" ) downto field_lo(c_hdr_field_arr, "eth_src_mac")) <= x"00228608" & B"000" & ID(7 downto 3) & RESIZE_UVEC(ID(2 downto 0), c_byte_w);

    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "ip_src_addr" ) downto field_lo(c_hdr_field_arr, "ip_src_addr")) <= x"0A63" & B"000" & ID(7 downto 3) & INCR_UVEC(RESIZE_UVEC(ID(2 downto 0), c_byte_w), 1);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "ip_dst_addr" ) downto field_lo(c_hdr_field_arr, "ip_dst_addr")) <= x"0A63" & B"000" & ID(7 downto 3) & INCR_UVEC(RESIZE_UVEC(ID(2 downto 0), c_byte_w), 1);

    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "usr_sync"    ) downto field_lo(c_hdr_field_arr, "usr_sync"   )) <= slv(dp_offload_tx_snk_in_arr(i).sync);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "usr_bsn"     ) downto field_lo(c_hdr_field_arr, "usr_bsn"    )) <= dp_offload_tx_snk_in_arr(i).bsn(59 downto 0);
  end generate;

  -----------------------------------------------------------------------------
  -- RX: dp_offload_rx
  -----------------------------------------------------------------------------
  u_dp_offload_rx : entity dp_lib.dp_offload_rx
  generic map (
    g_nof_streams         => c_nof_streams,
    g_data_w              => c_data_w,
    g_hdr_field_arr       => c_hdr_field_arr,
    g_remove_crc          => true,
    g_crc_nof_words       => 1
   )
  port map (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,

    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

    reg_hdr_dat_mosi      => reg_dp_offload_rx_hdr_dat_mosi,
    reg_hdr_dat_miso      => reg_dp_offload_rx_hdr_dat_miso,

    snk_in_arr            => dp_offload_rx_snk_in_arr,
    snk_out_arr           => dp_offload_rx_snk_out_arr,

    src_out_arr           => dp_offload_rx_src_out_arr,
    src_in_arr            => dp_offload_rx_src_in_arr,

    hdr_fields_out_arr    => hdr_fields_out_arr
    );

  gen_hdr_out_fields : for i in 0 to c_nof_streams - 1 generate
    diag_data_buf_snk_in_arr(i).sync <=          sl(hdr_fields_out_arr(i)(field_hi(c_hdr_field_arr, "usr_sync") downto field_lo(c_hdr_field_arr, "usr_sync" )));
    diag_data_buf_snk_in_arr(i).bsn  <= RESIZE_UVEC(hdr_fields_out_arr(i)(field_hi(c_hdr_field_arr, "usr_bsn" ) downto field_lo(c_hdr_field_arr, "usr_bsn"  )), c_dp_stream_bsn_w);
  end generate;

  -----------------------------------------------------------------------------
  -- RX: Data buffers and BSN monitors
  -----------------------------------------------------------------------------
  dp_offload_rx_src_in_arr <= diag_data_buf_snk_out_arr;

  gen_bsn_mon_in : for i in 0 to c_nof_streams - 1 generate
    diag_data_buf_snk_in_arr(i).data  <= dp_offload_rx_src_out_arr(i).data;
    diag_data_buf_snk_in_arr(i).valid <= dp_offload_rx_src_out_arr(i).valid;
    diag_data_buf_snk_in_arr(i).sop   <= dp_offload_rx_src_out_arr(i).sop;
    diag_data_buf_snk_in_arr(i).eop   <= dp_offload_rx_src_out_arr(i).eop;
    diag_data_buf_snk_in_arr(i).err   <= dp_offload_rx_src_out_arr(i).err;
  end generate;

  u_dp_bsn_monitor : entity dp_lib.mms_dp_bsn_monitor
  generic map (
    g_nof_streams        => c_nof_streams,
    g_cross_clock_domain => true,
    g_sync_timeout       => 2 * c_bg_blocks_per_sync * (c_bg_block_size + c_bg_gapsize),
    g_cnt_sop_w          => ceil_log2(c_bg_blocks_per_sync + 1),
    g_cnt_valid_w        => ceil_log2(c_bg_blocks_per_sync * c_bg_block_size+1),
    g_log_first_bsn      => true
  )
  port map (
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_mosi,
    reg_miso    => reg_bsn_monitor_miso,

    dp_rst      => dp_rst,
    dp_clk      => dp_clk,
    in_siso_arr => diag_data_buf_snk_out_arr,
    in_sosi_arr => diag_data_buf_snk_in_arr
  );

  diag_data_buf_snk_out_arr <= (others => c_dp_siso_rdy);

  u_diag_data_buffer : entity diag_lib.mms_diag_data_buffer
  generic map (
    g_nof_streams  => c_nof_streams,
    g_data_w       => c_data_w,
    g_buf_nof_data => 1024,
    g_buf_use_sync => true
  )
  port map (
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,

    ram_data_buf_mosi => ram_diag_data_buf_mosi,
    ram_data_buf_miso => ram_diag_data_buf_miso,
    reg_data_buf_mosi => reg_diag_data_buf_mosi,
    reg_data_buf_miso => reg_diag_data_buf_miso,

    in_sync           => diag_data_buf_snk_in_arr(0).sop,
    in_sosi_arr       => diag_data_buf_snk_in_arr
  );

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl_unb1_board : entity unb1_board_lib.ctrl_unb1_board
  generic map (
    g_sim                     => g_sim,
    g_design_name             => c_design_name,
    g_stamp_date              => g_stamp_date,
    g_stamp_time              => g_stamp_time,
    g_stamp_svn               => g_stamp_svn,
    g_mm_clk_freq             => c_unb1_board_mm_clk_freq_50M,
    g_udp_offload             => true,
    g_udp_offload_nof_streams => c_nof_streams,
    g_dp_clk_use_pll          => false
  )
  port map (
    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_locked                => mm_locked,
    mm_rst                   => mm_rst,

    dp_rst                   => dp_rst,
    dp_clk                   => OPEN,
    dp_pps                   => OPEN,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,

    -- PIOs
    pout_wdi                 => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    -- eth1g
    eth1g_tse_clk            => eth1g_tse_clk,
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- eth1g UDP streaming ports
    udp_tx_sosi_arr          =>  dp_offload_tx_src_out_arr,
    udp_tx_siso_arr          =>  dp_offload_tx_src_in_arr,
    udp_rx_sosi_arr          =>  dp_offload_rx_snk_in_arr,
    udp_rx_siso_arr          =>  dp_offload_rx_snk_out_arr,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso,
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso,

    -- UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,

    -- UniBoard FPGA pins
    CLK                      => '0',
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm_unb1_dp_offload : entity work.mmm_unb1_dp_offload
  generic map (
    g_sim           => g_sim,
    g_sim_unb_nr    => g_sim_unb_nr,
    g_sim_node_nr   => g_sim_node_nr,
    g_nof_streams   => c_nof_streams,
    g_bg_block_size => c_bg_block_size,
    g_hdr_field_arr => c_hdr_field_arr
   )
  port map(
    xo_clk                         => xo_clk,
    xo_rst_n                       => xo_rst_n,
    xo_rst                         => xo_rst,

    mm_rst                         => mm_rst,
    mm_clk                         => mm_clk,
    mm_locked                      => mm_locked,

    dp_clk                         => dp_clk,

    -- PIOs
    pout_wdi                       => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi                   => reg_wdi_mosi,
    reg_wdi_miso                   => reg_wdi_miso,

    -- system_info
    reg_unb_system_info_mosi       => reg_unb_system_info_mosi,
    reg_unb_system_info_miso       => reg_unb_system_info_miso,
    rom_unb_system_info_mosi       => rom_unb_system_info_mosi,
    rom_unb_system_info_miso       => rom_unb_system_info_miso,

    -- UniBoard I2C sensors
    reg_unb_sens_mosi              => reg_unb_sens_mosi,
    reg_unb_sens_miso              => reg_unb_sens_miso,

    -- eth1g
    eth1g_tse_clk                  => eth1g_tse_clk,
    eth1g_mm_rst                   => eth1g_mm_rst,
    eth1g_tse_mosi                 => eth1g_tse_mosi,
    eth1g_tse_miso                 => eth1g_tse_miso,
    eth1g_reg_mosi                 => eth1g_reg_mosi,
    eth1g_reg_miso                 => eth1g_reg_miso,
    eth1g_reg_interrupt            => eth1g_reg_interrupt,
    eth1g_ram_mosi                 => eth1g_ram_mosi,
    eth1g_ram_miso                 => eth1g_ram_miso,

    ram_diag_bg_mosi               => ram_diag_bg_mosi,
    ram_diag_bg_miso               => ram_diag_bg_miso,
    reg_diag_bg_mosi               => reg_diag_bg_mosi,
    reg_diag_bg_miso               => reg_diag_bg_miso,

    reg_dp_offload_tx_hdr_dat_mosi => reg_dp_offload_tx_hdr_dat_mosi,
    reg_dp_offload_tx_hdr_dat_miso => reg_dp_offload_tx_hdr_dat_miso,

    reg_dp_offload_rx_hdr_dat_mosi => reg_dp_offload_rx_hdr_dat_mosi,
    reg_dp_offload_rx_hdr_dat_miso => reg_dp_offload_rx_hdr_dat_miso,

    reg_bsn_monitor_mosi           => reg_bsn_monitor_mosi,
    reg_bsn_monitor_miso           => reg_bsn_monitor_miso,

    ram_diag_data_buf_mosi         => ram_diag_data_buf_mosi,
    ram_diag_data_buf_miso         => ram_diag_data_buf_miso,
    reg_diag_data_buf_mosi         => reg_diag_data_buf_mosi,
    reg_diag_data_buf_miso         => reg_diag_data_buf_miso
  );
end str;
