-------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . The dp_stream_verify verifies the stream of packets with counter data that
--   are generated by dp_stimuli_st.
-- Description:
--   The component can verify a stream:
--   . The sosi control fields are verified conform the bus specifications
--     eg. considering the RL, no missing eop, etc.
--   . The sosi data fields are verified based on their previous value under
--     the assumption that they contain incrementing data. Whether a field
--     is checked depends on verify_snk_in_enable.
--
--   The component also checks whether the stream is active at all. A
--   pulse in verify_expected_snk_in_evt triggers the verification of the
--   corresponding field in snk_in using the expected_snk_in as reference.
--
-- Usage:
-- . See tb_dp_example_no_dut for usage example
--

library IEEE, common_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_lfsr_sequences_pkg.all;
use common_lib.tb_common_pkg.all;
use work.dp_stream_pkg.all;
use work.tb_dp_pkg.all;

entity dp_stream_verify is
  generic (
    g_instance_nr         : natural := 0;
    -- flow control
    -- . use different random width for stimuli and for verify to have different random sequences
    g_random_w            : natural := 14;
    g_pulse_active        : natural := 1;
    g_pulse_period        : natural := 2;
    g_flow_control        : t_dp_flow_control_enum := e_active;  -- always active, random or pulse flow control
    -- initializations
    g_sync_period         : natural := 10;
    g_sync_offset         : natural := 7;
    g_snk_in_cnt_max      : t_dp_sosi_unsigned := c_dp_sosi_unsigned_rst;  -- default 0 is no wrap
    g_snk_in_cnt_gap      : t_dp_sosi_unsigned := c_dp_sosi_unsigned_ones;  -- default only accept increment +1
    -- specific
    g_in_dat_w            : natural := 32;
    g_pkt_len             : natural := 16
  );
  port (
    rst                        : in  std_logic;
    clk                        : in  std_logic;

    -- Verify data
    snk_out                    : out t_dp_siso;
    snk_in                     : in  t_dp_sosi;

    -- During stimuli
    verify_snk_in_enable       : in  t_dp_sosi_sl;  -- enable to verify that the snk_in fields are incrementing

    -- End of stimuli
    expected_snk_in            : in  t_dp_sosi;  -- expected snk_in at verify_expected_snk_in_evt
    verify_expected_snk_in_evt : in  t_dp_sosi_sl  -- trigger to verify the expected_snk_in
  );
end dp_stream_verify;

architecture tb of dp_stream_verify is
  constant c_rl                       : natural := 1;
  constant c_no_dut                   : boolean := true;

  -- Use different initialization by g_instance_nr, to have different random sequences per stream
  signal random                     : std_logic_vector(g_random_w - 1 downto 0) := TO_UVEC(g_instance_nr, g_random_w);
  signal pulse                      : std_logic;
  signal pulse_en                   : std_logic := '1';

  signal i_snk_out                  : t_dp_siso := c_dp_siso_rdy;
  signal prev_snk_out               : t_dp_siso;

  -- hold valid data for verify at verify_expected_snk_in_evt
  signal hold_snk_in_data           : std_logic_vector(c_dp_stream_data_w - 1 downto 0);
  signal hold_snk_in_re             : std_logic_vector(c_dp_stream_dsp_data_w - 1 downto 0);
  signal hold_snk_in_im             : std_logic_vector(c_dp_stream_dsp_data_w - 1 downto 0);

  signal snk_in_data                : std_logic_vector(g_in_dat_w - 1 downto 0);
  signal prev_snk_in                : t_dp_sosi;

  signal hold_snk_in_sop            : std_logic := '0';
  signal detected_snk_in_ctrl       : t_dp_sosi_sl := c_dp_sosi_sl_rst;
  signal verify_snk_in_increment    : t_dp_sosi_sl := c_dp_sosi_sl_rst;
  signal verify_snk_in_ctrl         : t_dp_sosi_sl := c_dp_sosi_sl_rst;

  signal exp_size                   : natural;
  signal cnt_size                   : natural;
begin
  snk_out <= i_snk_out;

  ------------------------------------------------------------------------------
  -- STREAM CONTROL
  ------------------------------------------------------------------------------

  random <= func_common_random(random) when rising_edge(clk);

  proc_common_gen_duty_pulse(g_pulse_active, g_pulse_period, '1', rst, clk, pulse_en, pulse);

  i_snk_out.ready <= '1'                 when g_flow_control = e_active  else
                     random(random'high) when g_flow_control = e_random  else
                     pulse               when g_flow_control = e_pulse;

  ------------------------------------------------------------------------------
  -- DATA VERIFICATION
  ------------------------------------------------------------------------------

  -- Detect first sync, sop, eop, valid
  detected_snk_in_ctrl.sync  <= '1' when snk_in.sync = '1'  and rising_edge(clk);
  detected_snk_in_ctrl.valid <= '1' when snk_in.valid = '1' and rising_edge(clk);
  detected_snk_in_ctrl.sop   <= '1' when snk_in.sop = '1'   and rising_edge(clk);
  detected_snk_in_ctrl.eop   <= '1' when snk_in.eop = '1'   and rising_edge(clk);

  -- Verify that the stimuli have been applied at all so at least one active sosi sync, sop, eop, valid field has been detected
  proc_dp_verify_value("snk_in.sync",             clk, verify_expected_snk_in_evt.sync,    expected_snk_in.sync,    detected_snk_in_ctrl.sync);
  proc_dp_verify_value("snk_in.sop",              clk, verify_expected_snk_in_evt.sop,     expected_snk_in.sop,     detected_snk_in_ctrl.sop);
  proc_dp_verify_value("snk_in.eop",              clk, verify_expected_snk_in_evt.eop,     expected_snk_in.eop,     detected_snk_in_ctrl.eop);
  proc_dp_verify_value("snk_in.valid",            clk, verify_expected_snk_in_evt.valid,   expected_snk_in.valid,   detected_snk_in_ctrl.valid);

  -- Verify that the last sosi data, re, im, bsn, channel and err fields are correct
  proc_dp_verify_value("snk_in.data",    e_equal, clk, verify_expected_snk_in_evt.data,    expected_snk_in.data,    hold_snk_in_data);
  proc_dp_verify_value("snk_in.re",      e_equal, clk, verify_expected_snk_in_evt.re,      expected_snk_in.re,      hold_snk_in_re);
  proc_dp_verify_value("snk_in.im",      e_equal, clk, verify_expected_snk_in_evt.im,      expected_snk_in.im,      hold_snk_in_im);
  proc_dp_verify_value("snk_in.bsn",     e_equal, clk, verify_expected_snk_in_evt.bsn,     expected_snk_in.bsn,     snk_in.bsn);
  proc_dp_verify_value("snk_in.channel", e_equal, clk, verify_expected_snk_in_evt.channel, expected_snk_in.channel, snk_in.channel);
  proc_dp_verify_value("snk_in.err",     e_equal, clk, verify_expected_snk_in_evt.err,     expected_snk_in.err,     snk_in.err);

  -- Verify that the output is incrementing data, like the input stimuli
  p_verify_snk_in_increment : process(verify_snk_in_enable, detected_snk_in_ctrl)
  begin
    verify_snk_in_increment         <= verify_snk_in_enable;
    verify_snk_in_increment.data    <= verify_snk_in_enable.data    and detected_snk_in_ctrl.valid;
    verify_snk_in_increment.re      <= verify_snk_in_enable.re      and detected_snk_in_ctrl.valid;
    verify_snk_in_increment.im      <= verify_snk_in_enable.im      and detected_snk_in_ctrl.valid;
    verify_snk_in_increment.bsn     <= verify_snk_in_enable.bsn     and detected_snk_in_ctrl.sop;
    verify_snk_in_increment.channel <= verify_snk_in_enable.channel and detected_snk_in_ctrl.sop;
    verify_snk_in_increment.empty   <= verify_snk_in_enable.empty   and detected_snk_in_ctrl.eop;
    verify_snk_in_increment.err     <= verify_snk_in_enable.err     and detected_snk_in_ctrl.eop;
  end process;

  proc_dp_verify_data("snk_in.data",    c_rl, g_snk_in_cnt_max.data,    g_snk_in_cnt_gap.data,    clk, verify_snk_in_increment.data,    i_snk_out.ready, snk_in.valid, snk_in.data,    prev_snk_in.data);
  proc_dp_verify_data("snk_in.re",      c_rl, g_snk_in_cnt_max.re,      g_snk_in_cnt_gap.re,      clk, verify_snk_in_increment.re,      i_snk_out.ready, snk_in.valid, snk_in.re,      prev_snk_in.re);
  proc_dp_verify_data("snk_in.im",      c_rl, g_snk_in_cnt_max.im,      g_snk_in_cnt_gap.im,      clk, verify_snk_in_increment.im,      i_snk_out.ready, snk_in.valid, snk_in.im,      prev_snk_in.im);
  proc_dp_verify_data("snk_in.bsn",     c_rl, g_snk_in_cnt_max.bsn,     g_snk_in_cnt_gap.bsn,     clk, verify_snk_in_increment.bsn,     i_snk_out.ready, snk_in.sop,   snk_in.bsn,     prev_snk_in.bsn);
  proc_dp_verify_data("snk_in.channel", c_rl, g_snk_in_cnt_max.channel, g_snk_in_cnt_gap.channel, clk, verify_snk_in_increment.channel, i_snk_out.ready, snk_in.sop,   snk_in.channel, prev_snk_in.channel);
  proc_dp_verify_data("snk_in.empty",   c_rl, g_snk_in_cnt_max.empty,   g_snk_in_cnt_gap.empty,   clk, verify_snk_in_increment.empty,   i_snk_out.ready, snk_in.eop,   snk_in.empty,   prev_snk_in.empty);
  proc_dp_verify_data("snk_in.err",     c_rl, g_snk_in_cnt_max.err,     g_snk_in_cnt_gap.err,     clk, verify_snk_in_increment.err,     i_snk_out.ready, snk_in.eop,   snk_in.err,     prev_snk_in.err);

  -- Verify that the snk_in control fields are correct
  p_verify_snk_in_ctrl: process(snk_in, verify_snk_in_enable)
  begin
    verify_snk_in_ctrl.sync  <= snk_in.sync  and verify_snk_in_enable.valid and verify_snk_in_enable.sync;
    verify_snk_in_ctrl.sop   <= snk_in.sop   and verify_snk_in_enable.valid and verify_snk_in_enable.sop and verify_snk_in_enable.eop;
    verify_snk_in_ctrl.eop   <= snk_in.eop   and verify_snk_in_enable.valid and verify_snk_in_enable.sop and verify_snk_in_enable.eop;
    verify_snk_in_ctrl.valid <= snk_in.valid and verify_snk_in_enable.valid;
  end process;

  -- Verify that the output sync occurs when expected
  proc_dp_verify_sync(g_sync_period, g_sync_offset, clk, detected_snk_in_ctrl.sop, verify_snk_in_ctrl.sync, verify_snk_in_ctrl.sop, snk_in.bsn);

  -- Verify output packet ctrl
  proc_dp_verify_sop_and_eop(clk, verify_snk_in_ctrl.valid, verify_snk_in_ctrl.sop, verify_snk_in_ctrl.eop, hold_snk_in_sop);

  -- Verify output packet block size
  exp_size <= g_pkt_len;

  proc_dp_verify_block_size(exp_size, clk, verify_snk_in_ctrl.valid, verify_snk_in_ctrl.sop, verify_snk_in_ctrl.eop, cnt_size);

  -- Verify output ready latency
  proc_dp_verify_valid(clk, detected_snk_in_ctrl.valid, i_snk_out.ready, prev_snk_out.ready, verify_snk_in_ctrl.valid);

  ------------------------------------------------------------------------------
  -- Auxiliary
  ------------------------------------------------------------------------------

  -- Map to slv to ease monitoring in wave window
  snk_in_data  <= snk_in.data(g_in_dat_w - 1 downto 0);

  hold_snk_in_data <= snk_in.data when snk_in.valid = '1';
  hold_snk_in_re   <= snk_in.re   when snk_in.valid = '1';
  hold_snk_in_im   <= snk_in.im   when snk_in.valid = '1';
end tb;
