-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: E. Kooistra
-- Purpose: Connect a single MM master interface to a list of MM slave
--          interfaces using mm_bus_pipe.
-- Description:
--   In addition to mm_bus_pipe this mm_bus takes care of:
--   - not connected slaves
--   - slaves that do not need flow control
--
--                                     FOR g_nof_slaves:
--                             g_slave_enable_arr
--                             g_waitrequest_arr
--                             g_rd_latency_arr
--                              |          |
--                              |          |
--        g_base_arr            |          |
--        g_width_arr           |          |
--        g_pipeline_mosi       |          |
--        g_pipeline_miso_rdval |          |
--        g_pipeline_miso_wait  |          |
--                          |   |          |
--                        __v___v_     ____v___
--   master_miso <-------| mm_bus |<--| mm     |<-- slave_miso_arr
--   master_mosi ------->| pipe   |-->| slave  |--> slave_mosi_arr
--                       |________|   | enable |
--                                    |________|
--   The mm_bus_comb takes care of:
--   - MM bus multiplexer between master and slaves
--   - MM slave address allocation on the MM bus
--   The mm_bus_pipe takes care of:
--   - Pipelining the MM bus, the mm_bus_comb and mm_slave_enable are
--     combinatorial.
--
-- Usage:
--   The ascii drawing shows how this mm_bus can be used in combination
--   with other MM bus components to create an memory mapped bus:
--
--   . mm_bus        : connects a master to multiple independent slaves
--   . mm_slave_mux  : connects an array of slave to a single slave port
--   . mm_master_mux : connects mulitple masters to a single slave
--
--                mm_slave_mux
--         mm_bus     |---S
--            |-------|---S
--            |---S   |---S
--       M ---|
--            |---S
--            |---S
--            |-------|
--            |---S   |---S
--                    |
--       M -----------|
--                mm_master_mux
--
--     The mm_slave_mux and mm_master_mux should typically be combinatorial,
--     because all MM bus pipelining is concentrated in mm_bus_pipe.
--
--   * The mm_slave_mux is useful to present an array of equal slave MM
--     ports via a single port on the MM bus. Otherwise the mm_bus could
--     instead directly present each slave MM array port.
--     The mm_slave_mux introduces hierarchy in the MM bus structure. This
--     can help to influcence the timing closure. Using only mm_bus or
--     the a combination of mm_bus and mm_slave_mux can help to steer
--     where pipelining is inserted in the MM bus.
--   * The MM bus based on mm_bus could be automatically generated by ARGS
--     based on a set of MM slave ports described in YAML configuration
--     files.
--
-- Limitations --> see mm_bus_comb
--
-- Todo (only if necessary):
-- * The mm_bus assumes that the MM slave will eventually acknowledge an
--   mosi.wr or rd access by pulling miso.waitrequest low. If the MM slave
--   malfunctions then the MM bus access will stall. Therefore a MM slave
--   port that uses mosi flow control should also support a waitrequest
--   timeout mechanism. Such a waitrequest timeout mechanism could be made
--   part of mm_slave_enable.
-- * The master can then be informed about a failing mosi access using
--   the miso.response field of the Avalon bus. A failing mosi access can
--   be due to a not connected slave, an out-of-range address, a slave that
--   times out on flow control.
--
-------------------------------------------------------------------------------

library IEEE, common_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use common_lib.common_mem_pkg.all;

entity mm_bus is
  generic (
    g_nof_slaves          : positive;  -- Number of MM slave interfaces on the bus
    g_base_arr            : t_nat_natural_arr;  -- Address base per slave
    g_width_arr           : t_nat_natural_arr;  -- Address width per slave
    g_rd_latency_arr      : t_nat_natural_arr;  -- Read latency per slave
    g_slave_enable_arr    : t_nat_boolean_arr;  -- Use FALSE for not connected slaves, else TRUE
    g_waitrequest_arr     : t_nat_boolean_arr;  -- Enable waitrequest flow control per slave, else fixed '0'
    g_pipeline_mosi       : boolean := false;  -- Pipeline MM access (wr, rd)
    g_pipeline_miso_rdval : boolean := false;  -- Pipeline MM read (rdval)
    g_pipeline_miso_wait  : boolean := false  -- Pipeline MM access flow control (waitrequest)
  );
  port (
    mm_rst         : in  std_logic := '0';
    mm_clk         : in  std_logic := '0';
    master_mosi    : in  t_mem_mosi;
    master_miso    : out t_mem_miso;
    slave_mosi_arr : out t_mem_mosi_arr(0 to g_nof_slaves - 1);
    slave_miso_arr : in  t_mem_miso_arr(0 to g_nof_slaves - 1) := (others => c_mem_miso_rst)
  );
end mm_bus;

architecture str of mm_bus is
  signal bus_mosi_arr   : t_mem_mosi_arr(0 to g_nof_slaves - 1);
  signal bus_miso_arr   : t_mem_miso_arr(0 to g_nof_slaves - 1);
begin
  -- MM bus
  u_mm_bus_pipe : entity work.mm_bus_pipe
  generic map (
    g_nof_slaves          => g_nof_slaves,
    g_base_arr            => g_base_arr,
    g_width_arr           => g_width_arr,
    g_rd_latency_arr      => g_rd_latency_arr,
    g_waitrequest_arr     => g_waitrequest_arr,
    g_pipeline_mosi       => g_pipeline_mosi,
    g_pipeline_miso_rdval => g_pipeline_miso_rdval,
    g_pipeline_miso_wait  => g_pipeline_miso_wait
  )
  port map (
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    master_mosi    => master_mosi,
    master_miso    => master_miso,
    slave_mosi_arr => bus_mosi_arr,
    slave_miso_arr => bus_miso_arr
  );

  -- The MM bus interface with the MM slaves
  gen_slave_ports : for I in 0 to g_nof_slaves - 1 generate
    -- Rewire not connected slaves and slave that do not need mosi flow control via miso.waitrequest
    u_slave_enable : entity work.mm_slave_enable
    generic map (
      g_enable       => g_slave_enable_arr(I),
      g_waitrequest  => g_waitrequest_arr(I),
      g_rd_latency   => g_rd_latency_arr(I)
    )
    port map (
      mm_rst        => mm_rst,
      mm_clk        => mm_clk,
      -- MM input RL = 1
      in_mosi       => bus_mosi_arr(I),
      in_miso       => bus_miso_arr(I),
      -- MM output RL = 0
      out_mosi      => slave_mosi_arr(I),
      out_miso      => slave_miso_arr(I)
    );
  end generate;

end str;
