-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
--     Jonathan Hargreaves, 18 Nov 2016
--
-- Purpose:
--     Test bench for iquv.vhd
-- Description:
--     Test bench is self checking using reference data generated by iquv_checker.py
-- Usage:
--     do wave_iquv.do
--     run -all

library IEEE, common_lib, dp_lib, technology_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_lfsr_sequences_pkg.all;
use dp_lib.dp_stream_pkg.all;
use technology_lib.technology_select_pkg.all;

entity tb_iquv is
  generic (
    g_nof_int          : natural := 16;  -- number of successive subbands to integrate
    g_random_valid     : boolean := true;  -- Put gaps in the valid signal
    g_in_data_w        : natural := 12;  -- width of the incoming complex data
    g_out_data_w       : natural := 18  -- width of the output data
  );
end tb_iquv;

architecture tb of tb_iquv is
  constant clk_period    : time := 10 ns;
  constant c_npoints     : natural := 4 * g_nof_int;  -- number of subbands (size of block) at present only used to generate test vectors
  constant c_fsd_w       : natural := g_in_data_w * 2 + 1 + ceil_log2(g_nof_int);  -- The number of bits used for the accumulated result inside the DUT
  constant c_lsb_w       : natural := c_fsd_w - g_out_data_w;  -- number of lsbs to remove from the expected result

  signal tb_end          : std_logic := '0';
  signal dp_rst          : std_logic := '1';
  signal dp_clk          : std_logic := '0';

  signal in_complex      : t_dp_sosi;
  signal i_out           : t_dp_sosi;
  signal q_out           : t_dp_sosi;
  signal u_out           : t_dp_sosi;
  signal v_out           : t_dp_sosi;

  signal diag_i_out_data : std_logic_vector(31 downto 0);
  signal diag_q_out_data : std_logic_vector(31 downto 0);
  signal diag_u_out_data : std_logic_vector(31 downto 0);
  signal diag_v_out_data : std_logic_vector(31 downto 0);
  signal diag_out_valid  : std_logic;

  signal random0         : std_logic_vector(7 downto 0) := (others => '0');
  signal valid_enable    : std_logic := '0';

  -- expected data values
  constant i_nint8_exp_arr : t_integer_arr(0 to 35) := (140,1100,3084,6092,
                                                     140,1100,3084,6092,
                                                     140,1100,3084,6092,
                                                     140,1100,3084,6092,
                                                     280,2200,6168,12184,
                                                     280,2200,6168,12184,
                                                     560,4400,12336,24368,
                                                     84864,84864,84864,84864,
                                                     134086688,134086688,134086688,134086688);
  constant q_nint8_exp_arr : t_integer_arr(0 to 35) := (140,1100,3084,6092,
                                                     - 140, -1100, -3084, -6092,
                                                     140,1100,3084,6092,
                                                     - 140, -1100, -3084, -6092,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     - 77440, -77440, -77440, -77440,
                                                     0,0,0,0);
  constant u_nint8_exp_arr : t_integer_arr(0 to 35) := (0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     280,2200,6168,12184,
                                                     280,2200,6168,12184,
                                                     560,4400,12336,24368,
                                                     - 16640, -16640, -16640, -16640,
                                                     - 134086688, -134086688, -134086688, -134086688);
  constant v_nint8_exp_arr : t_integer_arr(0 to 35) := (0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     - 30464, -30464, -30464, -30464,
                                                     0,0,0,0);

  constant i_nint16_exp_arr : t_integer_arr(0 to 35) := (1240,9176,25304,49624,
                                                     1240,9176,25304,49624,
                                                     1240,9176,25304,49624,
                                                     1240,9176,25304,49624,
                                                     2480,18352,50608,99248,
                                                     2480,18352,50608,99248,
                                                     4960,36704,101216,198496,
                                                     169728,169728,169728,169728,
                                                     268173376,268173376,268173376,268173376);
  constant q_nint16_exp_arr : t_integer_arr(0 to 35) := (1240,9176,25304,49624,
                                                     - 1240, -9176, -25304, -49624,
                                                     1240,9176,25304,49624,
                                                     - 1240, -9176, -25304, -49624,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     - 154880, -154880, -154880, -154880,
                                                     0,0,0,0);
  constant u_nint16_exp_arr : t_integer_arr(0 to 35) := (0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     2480,18352,50608,99248,
                                                     2480,18352,50608,99248,
                                                     4960,36704,101216,198496,
                                                     - 33280, -33280, -33280, -33280,
                                                     - 268173376, -268173376, -268173376, -268173376);
  constant v_nint16_exp_arr : t_integer_arr(0 to 35) := (0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     0,0,0,0,
                                                     - 60928, -60928, -60928, -60928,
                                                     0,0,0,0);
begin
  dp_clk  <= not dp_clk or tb_end after clk_period / 2;
  dp_rst  <= '0' after clk_period * 10;
  random0 <= func_common_random(random0) when rising_edge(dp_clk);
  valid_enable <= random0(random0'high) when g_random_valid else '1';

  ------------------------------------------------------------------------------
  -- Valid signal stimuli
  ------------------------------------------------------------------------------

  p_valstimuli : process
  begin
    in_complex.valid <= '0';
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;
    while true loop
      --in_complex.valid <= '1';
      in_complex.valid <= valid_enable;
      wait until rising_edge(dp_clk);
    end loop;
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- Input stimuli
  ------------------------------------------------------------------------------

  p_stimuli : process
  begin
    in_complex.re <= TO_DP_DSP_DATA(0);
    in_complex.im <= TO_DP_DSP_DATA(0);
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;

    for BLOCKCOUNT in 0 to 0 loop  -- Repeat as needed

      in_complex.im <= TO_DP_DSP_DATA(0);  -- Keep the imaginary part 0
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle one
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the X pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(0);  -- The real part of the Y pol stays 0
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle two
        in_complex.re <= TO_DP_DSP_DATA(0);  -- The real part of the X pol stays 0
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the Y pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;

      in_complex.re <= TO_DP_DSP_DATA(0);  -- Keep the real part 0
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle three
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the X pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.im <= TO_DP_DSP_DATA(0);  -- The im part of the Y pol stays 0
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle four
        in_complex.im <= TO_DP_DSP_DATA(0);  -- The im part of the X pol stays 0
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the Y pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;

      in_complex.im <= TO_DP_DSP_DATA(0);  -- Keep the imaginary part 0
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle five
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the X pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the Y pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;
      in_complex.re <= TO_DP_DSP_DATA(0);  -- Keep the real part 0
      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle six
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the X pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the Y pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;

      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle seven
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the X pol counts up
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the X pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(FBIN);  -- The real part of the Y pol counts up
        in_complex.im <= TO_DP_DSP_DATA(FBIN);  -- The im part of the Y pol counts up
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;

      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle eight
        in_complex.re <= TO_DP_DSP_DATA(-8);  -- The real part of the X pol is fixed
        in_complex.im <= TO_DP_DSP_DATA(20);  -- The im part of the X pol is fixed
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(100);  -- The real part of the Y pol is fixed
        in_complex.im <= TO_DP_DSP_DATA(-12);  -- The im part of the Y pol is fixed
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;

      for FBIN in 0 to c_npoints - 1 loop  -- BLOCK cycle nine
        in_complex.re <= TO_DP_DSP_DATA(2047);  -- The real part of the X pol is fixed
        in_complex.im <= TO_DP_DSP_DATA(2047);  -- The im part of the X pol is fixed
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(-2047);  -- The real part of the Y pol is fixed
        in_complex.im <= TO_DP_DSP_DATA(-2047);  -- The im part of the Y pol is fixed
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;
    end loop;

    for I in 0 to 9 loop wait until rising_edge(dp_clk); end loop;
    tb_end <= '1';
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------

  dut : entity work.iquv
  generic map (
    g_sim              => true,
    g_technology       => c_tech_select_default,
    g_use_accum        => true,
    g_nof_int          => g_nof_int,
    g_in_data_w        => g_in_data_w,
    g_out_data_w       => g_out_data_w
  )
  port map (
    dp_rst     => dp_rst,
    dp_clk     => dp_clk,
    in_complex => in_complex,
    i_out      => i_out,
    q_out      => q_out,
    u_out      => u_out,
    v_out      => v_out
  );

  -- diag signals to make outputs visible in Modelsim wave

  diag_i_out_data <= i_out.data(c_word_w - 1 downto 0);
  diag_q_out_data <= q_out.data(c_word_w - 1 downto 0);
  diag_u_out_data <= u_out.data(c_word_w - 1 downto 0);
  diag_v_out_data <= v_out.data(c_word_w - 1 downto 0);
  diag_out_valid  <= i_out.valid;

  p_verify : process(dp_clk, diag_out_valid, diag_i_out_data, diag_q_out_data, diag_u_out_data, diag_v_out_data)
  variable v_index : integer := 0;
  begin
    if rising_edge(dp_clk) then
      if diag_out_valid = '1' then
        if g_nof_int = 8 then
          assert TO_UINT(i_out.data) = TO_UINT(u_round(TO_UVEC(i_nint8_exp_arr(v_index),c_fsd_w - 1), c_lsb_w - 1, true)) report "Error: wrong result in I out DUT" severity ERROR;
          --REPORT "I exp = " & integer'image(to_uint(s_round(TO_UVEC(i_nint8_exp_arr(v_index),c_fsd_w-1), c_lsb_w-1, TRUE)));
          assert TO_SINT(q_out.data) = TO_SINT(s_round(TO_SVEC(q_nint8_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in Q out DUT" severity ERROR;
          assert TO_SINT(u_out.data) = TO_SINT(s_round(TO_SVEC(u_nint8_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in U out DUT" severity ERROR;
          assert TO_SINT(v_out.data) = TO_SINT(s_round(TO_SVEC(v_nint8_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in V out DUT" severity ERROR;

        elsif g_nof_int = 16 then
          assert TO_UINT(i_out.data) = TO_UINT(u_round(TO_UVEC(i_nint16_exp_arr(v_index),c_fsd_w - 1), c_lsb_w - 1, true)) report "Error: wrong result in I out DUT" severity ERROR;
          --REPORT "I expected = " & integer'image(to_uint(u_round(TO_UVEC(i_nint16_exp_arr(v_index),c_fsd_w-1), c_lsb_w-1, TRUE)));
          assert TO_SINT(q_out.data) = TO_SINT(s_round(TO_SVEC(q_nint16_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in Q out DUT" severity ERROR;
          --REPORT "Q expected = " & integer'image(to_sint(s_round(TO_SVEC(q_nint16_exp_arr(v_index),c_word_w), c_lsb_w)));
          assert TO_SINT(u_out.data) = TO_SINT(s_round(TO_SVEC(u_nint16_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in U out DUT" severity ERROR;
          assert TO_SINT(v_out.data) = TO_SINT(s_round(TO_SVEC(v_nint16_exp_arr(v_index),c_word_w), c_lsb_w)) report "Error: wrong result in V out DUT" severity ERROR;
        end if;

        report "I = " & integer'image(to_uint(i_out.data)) & ", Q = " & integer'image(to_sint(q_out.data)) &
             ", U = " & integer'image(to_sint(u_out.data)) & ", V = " & integer'image(to_sint(v_out.data));
        v_index := v_index + 1;
      end if;
    end if;
  end process;
end tb;
