-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
--     Jonathan Hargreaves, 18 Nov 2016
--
-- Purpose:
--     Test bench for iquv.vhd
-- Description:
--     Test bench is self checking using reference data generated by iquv_testgen.py and stored in g_data_file
-- Usage:
--     as 4
--     run -all

library IEEE, common_lib, dp_lib, technology_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.tb_common_pkg.all;
use common_lib.common_lfsr_sequences_pkg.all;
use dp_lib.dp_stream_pkg.all;
use technology_lib.technology_select_pkg.all;

entity tb_iquv_file_data is
  generic (
    g_random_valid     : boolean := true;  -- Put gaps in the valid signal
    g_data_file        : string  := "iquv_stimuli_1.dat";
    g_use_accum        : boolean := true;  -- Use the accumulator stage
    g_nof_int          : natural := 16;  -- Number of samples to integrate when accumulator is used
    g_in_data_w        : natural := 12;  -- width of the incoming complex data
    g_out_data_w       : natural := 18  -- width of the output data
  );
end tb_iquv_file_data;

architecture tb of tb_iquv_file_data is
  constant clk_period    : time := 10 ns;
  constant c_fsd_w       : natural := g_in_data_w * 2 + 1 + true_log2(g_nof_int);  -- The number of bits used for the accumulated result inside the DUT
  constant c_lsb_w       : natural := c_fsd_w - g_out_data_w;  -- number of lsbs to remove from the expected result

  signal tb_end          : std_logic := '0';
  signal dp_rst          : std_logic := '1';
  signal dp_clk          : std_logic := '0';

  signal in_complex      : t_dp_sosi;
  signal i_out           : t_dp_sosi;
  signal q_out           : t_dp_sosi;
  signal u_out           : t_dp_sosi;
  signal v_out           : t_dp_sosi;

  signal diag_i_out_data : std_logic_vector(31 downto 0);
  signal diag_q_out_data : std_logic_vector(31 downto 0);
  signal diag_u_out_data : std_logic_vector(31 downto 0);
  signal diag_v_out_data : std_logic_vector(31 downto 0);
  signal diag_out_valid  : std_logic;

  signal random0         : std_logic_vector(7 downto 0) := (others => '0');
  signal valid_enable    : std_logic := '0';

  -- Data file access
  constant c_nof_nof_int_cases      : natural := 3;
  constant c_nof_int_arr            : t_integer_arr(0 to c_nof_nof_int_cases - 1) := (8, 16, 1);  -- The nof_int cases to test. 1 means no accumulator
  constant c_nof_cols               : natural := 4 + 4 * c_nof_nof_int_cases;
  constant c_nof_lines              : natural := 14641;  -- 11 ** 4

  signal file_data_arr              : t_integer_arr(0 to (c_nof_lines * c_nof_cols) - 1) := (others => 0);
  signal xreal_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal ximag_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal yreal_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal yimag_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);

  signal i_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal i_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal i_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
begin
  dp_clk  <= not dp_clk or tb_end after clk_period / 2;
  dp_rst  <= '0' after clk_period * 10;
  random0 <= func_common_random(random0) when rising_edge(dp_clk);
  valid_enable <= random0(random0'high) when g_random_valid else '1';

  ------------------------------------------------------------------------------
  -- Valid signal stimuli
  ------------------------------------------------------------------------------

  p_valstimuli : process
  begin
    in_complex.valid <= '0';
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;
    while true loop
      --in_complex.valid <= '1';
      in_complex.valid <= valid_enable;
      wait until rising_edge(dp_clk);
    end loop;
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- Read stimuli and expected results from file
  ------------------------------------------------------------------------------

  p_stimuli : process
  begin
    proc_common_read_integer_file(g_data_file, 0, c_nof_lines, c_nof_cols, file_data_arr);
    wait for 1ns;
    for lineno in 0 to c_nof_lines - 1 loop
      xreal_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 0);
      ximag_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 1);
      yreal_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 2);
      yimag_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 3);
      i_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 4);
      q_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 5);
      u_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 6);
      v_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 7);
      i_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 8);
      q_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 9);
      u_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 10);
      v_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 11);
      i_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 12);
      q_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 13);
      u_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 14);
      v_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 15);
    end loop;
    in_complex.re <= TO_DP_DSP_DATA(0);
    in_complex.im <= TO_DP_DSP_DATA(0);
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;

    -- Apply the signals for the case g_nof_ints = 8
    if g_use_accum = true and g_nof_int = c_nof_int_arr(0) then
      for index in 0 to c_nof_lines - 1 loop
        for ii in 0 to c_nof_int_arr(0) - 1 loop
          in_complex.re <= TO_DP_DSP_DATA(xreal_stimuli_arr(index));
          in_complex.im <= TO_DP_DSP_DATA(ximag_stimuli_arr(index));
          wait until rising_edge(dp_clk) and valid_enable = '1';
          in_complex.re <= TO_DP_DSP_DATA(yreal_stimuli_arr(index));
          in_complex.im <= TO_DP_DSP_DATA(yimag_stimuli_arr(index));
          wait until rising_edge(dp_clk) and valid_enable = '1';
        end loop;
      end loop;
    -- Apply the signals for the case g_nof_ints = 16
    elsif g_use_accum = true and g_nof_int = c_nof_int_arr(1) then
      for index in 0 to c_nof_lines - 1 loop
        for ii in 0 to c_nof_int_arr(1) - 1 loop
          in_complex.re <= TO_DP_DSP_DATA(xreal_stimuli_arr(index));
          in_complex.im <= TO_DP_DSP_DATA(ximag_stimuli_arr(index));
          wait until rising_edge(dp_clk) and valid_enable = '1';
          in_complex.re <= TO_DP_DSP_DATA(yreal_stimuli_arr(index));
          in_complex.im <= TO_DP_DSP_DATA(yimag_stimuli_arr(index));
          wait until rising_edge(dp_clk) and valid_enable = '1';
        end loop;
      end loop;
    -- Apply the signals for the case no accumulator
    elsif g_use_accum = false then
      for index in 0 to c_nof_lines - 1 loop
        in_complex.re <= TO_DP_DSP_DATA(xreal_stimuli_arr(index));
        in_complex.im <= TO_DP_DSP_DATA(ximag_stimuli_arr(index));
        wait until rising_edge(dp_clk) and valid_enable = '1';
        in_complex.re <= TO_DP_DSP_DATA(yreal_stimuli_arr(index));
        in_complex.im <= TO_DP_DSP_DATA(yimag_stimuli_arr(index));
        wait until rising_edge(dp_clk) and valid_enable = '1';
      end loop;
    end if;

    for I in 0 to 9 loop wait until rising_edge(dp_clk); end loop;
    tb_end <= '1';
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------

  dut : entity work.iquv
  generic map (
    g_sim              => true,
    g_technology       => c_tech_select_default,
    g_use_accum        => g_use_accum,
    g_nof_int          => g_nof_int,
    g_in_data_w        => g_in_data_w,
    g_out_data_w       => g_out_data_w
  )
  port map (
    dp_rst     => dp_rst,
    dp_clk     => dp_clk,
    in_complex => in_complex,
    i_out      => i_out,
    q_out      => q_out,
    u_out      => u_out,
    v_out      => v_out
  );

  -- diag signals to make outputs visible in Modelsim wave

  diag_i_out_data <= i_out.data(c_word_w - 1 downto 0);
  diag_q_out_data <= q_out.data(c_word_w - 1 downto 0);
  diag_u_out_data <= u_out.data(c_word_w - 1 downto 0);
  diag_v_out_data <= v_out.data(c_word_w - 1 downto 0);
  diag_out_valid  <= i_out.valid;

  ------------------------------------------------------------------------------
  -- Results Checker
  ------------------------------------------------------------------------------
  p_verify : process(dp_clk, diag_out_valid, diag_i_out_data, diag_q_out_data, diag_u_out_data, diag_v_out_data)
  variable v_index : integer := 0;
  begin
    if rising_edge(dp_clk) then
      if diag_out_valid = '1' then
        if g_use_accum = true and g_nof_int = 8 then
          assert TO_UINT(i_out.data) = TO_UINT(u_round(TO_UVEC(i_nint8_exp_arr(v_index),c_fsd_w - 1), c_lsb_w - 1, true))
            report "Error: wrong result in I out DUT" severity ERROR;
          assert TO_SINT(q_out.data) = TO_SINT(s_round(TO_SVEC(q_nint8_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in Q out DUT" severity ERROR;
          assert TO_SINT(u_out.data) = TO_SINT(s_round(TO_SVEC(u_nint8_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in U out DUT" severity ERROR;
          assert TO_SINT(v_out.data) = TO_SINT(s_round(TO_SVEC(v_nint8_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in V out DUT" severity ERROR;
          --REPORT "I exp = " & integer'image(to_uint(s_round(TO_UVEC(i_nint8_exp_arr(v_index),c_fsd_w-1), c_lsb_w-1, TRUE)));

        elsif g_use_accum = true and g_nof_int = 16 then
          assert TO_UINT(i_out.data) = TO_UINT(u_round(TO_UVEC(i_nint16_exp_arr(v_index),c_fsd_w - 1), c_lsb_w - 1, true))
            report "Error: wrong result in I out DUT" severity ERROR;
          assert TO_SINT(q_out.data) = TO_SINT(s_round(TO_SVEC(q_nint16_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in Q out DUT" severity ERROR;
          assert TO_SINT(u_out.data) = TO_SINT(s_round(TO_SVEC(u_nint16_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in U out DUT" severity ERROR;
          assert TO_SINT(v_out.data) = TO_SINT(s_round(TO_SVEC(v_nint16_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in V out DUT" severity ERROR;
          --REPORT "I expected = " & integer'image(to_uint(u_round(TO_UVEC(i_nint16_exp_arr(v_index),c_fsd_w-1), c_lsb_w-1, TRUE)));
          --REPORT "Q expected = " & integer'image(to_sint(s_round(TO_SVEC(q_nint16_exp_arr(v_index),c_word_w), c_lsb_w)));

        elsif g_use_accum = false then
          assert TO_UINT(i_out.data) = TO_UINT(u_round(TO_UVEC(i_noint_exp_arr(v_index),c_fsd_w - 1), c_lsb_w - 1, true))
            report "Error: wrong result in I out DUT" severity ERROR;
          assert TO_SINT(q_out.data) = TO_SINT(s_round(TO_SVEC(q_noint_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in Q out DUT" severity ERROR;
          assert TO_SINT(u_out.data) = TO_SINT(s_round(TO_SVEC(u_noint_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in U out DUT" severity ERROR;
          assert TO_SINT(v_out.data) = TO_SINT(s_round(TO_SVEC(v_noint_exp_arr(v_index),c_fsd_w), c_lsb_w, true))
            report "Error: wrong result in V out DUT" severity ERROR;
          --REPORT "U expected = " & integer'image(to_sint(s_round(TO_SVEC(u_noint_exp_arr(v_index),c_fsd_w), c_lsb_w, TRUE)));
          --REPORT "U actual   = " & integer'image(to_sint(u_out.data));
        end if;

        --REPORT "I = " & integer'image(to_uint(i_out.data)) & ", Q = " & integer'image(to_sint(q_out.data)) &
        --     ", U = " & integer'image(to_sint(u_out.data)) & ", V = " & integer'image(to_sint(v_out.data));
        if v_index < c_nof_lines - 1 then
          v_index := v_index + 1;
        end if;
      end if;
    end if;
  end process;
end tb;
