-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
--     Jonathan Hargreaves, 12 Dec 2016
--
-- Purpose:
--     Test bench for iquv_iab.vhd
-- Description:
--     Test bench is self checking using reference data generated by iquv_testgen.py and stored in g_data_file
-- Usage:
--     as 4
--     run -all

library IEEE, common_lib, dp_lib, technology_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.tb_common_pkg.all;
use common_lib.common_lfsr_sequences_pkg.all;
use dp_lib.dp_stream_pkg.all;
use technology_lib.technology_select_pkg.all;

entity tb_iquv_iab_file_data is
  generic (
    g_random_valid     : boolean := true;  -- Put gaps in the valid signal
    g_data_file        : string  := "iquv_stimuli_1.dat";
    g_use_accum        : boolean := true;  -- Use the accumulator stage after the IAB
    g_nof_streams      : natural := 12;  -- number of imput streams
    g_nof_int          : natural := 16;  -- Number of samples to integrate when accumulator is used
    g_in_data_w        : natural := 12;  -- width of the incoming complex data
    g_out_data_w       : natural := 18  -- width of the output data
  );
end tb_iquv_iab_file_data;

architecture tb of tb_iquv_iab_file_data is
  constant clk_period    : time := 10 ns;
  constant c_out_iquv_w  : natural := 18;  -- number of bits the IQUV module truncates to
  constant c_iquv_accum_w: natural := sel_a_b(g_use_accum, 0, true_log2(g_nof_int));  -- Bitgrowth of the IQUV accumulator
  constant c_iab_accum_w : natural := sel_a_b(g_use_accum, true_log2(g_nof_int), 0);  -- Bitgrowth of the IAB accumulator
  constant c_fsd_iquv_w  : natural := g_in_data_w * 2 + 1 + c_iquv_accum_w;  -- The number of bits used for the result inside the IQUV
  constant c_lsb_iquv_w  : natural := c_fsd_iquv_w - c_out_iquv_w;  -- number of lsbs to remove from the expected IQUV result
  constant c_fsd_w       : natural := c_out_iquv_w + ceil_log2(g_nof_streams) + c_iab_accum_w;  -- The number of bits fsd after the IAB adder tree
                                                                                               -- and IAB accumulator if used
  constant c_lsb_w       : natural := c_fsd_w - g_out_data_w;  -- number of lsbs to remove from the expected IAB result
  constant c_allstreams  : natural := g_nof_streams;
  constant latency       : time := 100 ns;

  constant c_nof_iab_int : natural := sel_a_b(g_use_accum, g_nof_int, 1);  -- Used to simulate the IAB accumulator

  signal tb_end          : std_logic := '0';
  signal dp_rst          : std_logic := '1';
  signal dp_clk          : std_logic := '0';

  signal in_complex_data : t_dp_sosi := c_dp_sosi_rst;
  signal in_complex_zero : t_dp_sosi := c_dp_sosi_rst;
  signal in_complex_arr  : t_dp_sosi_arr(0 to g_nof_streams - 1);
  signal i_out           : t_dp_sosi;
  signal q_out           : t_dp_sosi;
  signal u_out           : t_dp_sosi;
  signal v_out           : t_dp_sosi;

  signal diag_i_out_data : std_logic_vector(31 downto 0);
  signal diag_q_out_data : std_logic_vector(31 downto 0);
  signal diag_u_out_data : std_logic_vector(31 downto 0);
  signal diag_v_out_data : std_logic_vector(31 downto 0);
  signal diag_out_valid  : std_logic;

  signal random0         : std_logic_vector(7 downto 0) := (others => '0');
  signal valid_enable    : std_logic := '0';

  -- Data file access
  constant c_nof_nof_int_cases    : natural := 3;
  constant c_nof_int_arr          : t_integer_arr(0 to c_nof_nof_int_cases - 1) := (8, 16, 1);  -- The nof_int cases to test. 1 means no accumulator
  constant c_nof_cols             : natural := 4 + 4 * c_nof_nof_int_cases;
  constant c_nof_lines            : natural := 14641;  -- 11 ** 4

  signal file_data_arr            : t_integer_arr(0 to (c_nof_lines * c_nof_cols) - 1) := (others => 0);
  signal xreal_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal ximag_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal yreal_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal yimag_stimuli_arr        : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);

  signal i_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_nint8_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal i_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_nint16_exp_arr         : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal i_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal q_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal u_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);
  signal v_noint_exp_arr          : t_integer_arr(0 to c_nof_lines - 1) := (others => 0);

  shared variable activestream    : natural := 0;
  signal activestream_dly: natural := 0;
begin
  dp_clk  <= not dp_clk or tb_end after clk_period / 2;
  dp_rst  <= '0' after clk_period * 10;
  random0 <= func_common_random(random0) when rising_edge(dp_clk);
  valid_enable <= random0(random0'high) when g_random_valid else '1';

  ------------------------------------------------------------------------------
  -- Copy input signals
  ------------------------------------------------------------------------------
  gen_copy_stimuli : for inputstream in 0 to g_nof_streams - 1 generate
    in_complex_arr(inputstream) <= in_complex_data when activestream = inputstream or activestream = c_allstreams else in_complex_zero;
  end generate;

  ------------------------------------------------------------------------------
  -- Valid signal stimuli
  ------------------------------------------------------------------------------

  p_valstimuli : process
  begin
    in_complex_data.valid <= '0';
    in_complex_zero.valid <= '0';
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;
    while true loop
      in_complex_data.valid <= valid_enable;
      in_complex_zero.valid <= valid_enable;
      wait until rising_edge(dp_clk);
    end loop;
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- Read stimuli and expected results from file
  ------------------------------------------------------------------------------

  p_stimuli : process
  begin
    proc_common_read_integer_file(g_data_file, 0, c_nof_lines, c_nof_cols, file_data_arr);
    wait for 1ns;
    for lineno in 0 to c_nof_lines - 1 loop
      xreal_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 0);
      ximag_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 1);
      yreal_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 2);
      yimag_stimuli_arr(lineno) <= file_data_arr(lineno * c_nof_cols + 3);
      i_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 4);
      q_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 5);
      u_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 6);
      v_nint8_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 7);
      i_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 8);
      q_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 9);
      u_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 10);
      v_nint16_exp_arr(lineno)  <= file_data_arr(lineno * c_nof_cols + 11);
      i_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 12);
      q_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 13);
      u_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 14);
      v_noint_exp_arr(lineno)   <= file_data_arr(lineno * c_nof_cols + 15);
    end loop;
    in_complex_data.re <= TO_DP_DSP_DATA(0);
    in_complex_data.im <= TO_DP_DSP_DATA(0);
    wait until rising_edge(dp_clk);
    for I in 0 to 19 loop wait until rising_edge(dp_clk); end loop;
    activestream := 0;
    -- Apply the stimuli to each stream in turn, then to all the streams
    for istream in 0 to g_nof_streams loop
      activestream := istream;
      if activestream > 0 and activestream < g_nof_streams then
      --IF activestream = 0 OR activestream = 1 THEN -- example for testing the next
        report "skipping activestream " & integer'image(activestream);
        next;
      else
        report "processing activestream " & integer'image(activestream);
        for index in 0 to c_nof_lines - 1 loop
          for ii in 0 to g_nof_int - 1 loop
            in_complex_data.re <= TO_DP_DSP_DATA(xreal_stimuli_arr(index));
            in_complex_data.im <= TO_DP_DSP_DATA(ximag_stimuli_arr(index));
            wait until rising_edge(dp_clk) and valid_enable = '1';
            in_complex_data.re <= TO_DP_DSP_DATA(yreal_stimuli_arr(index));
            in_complex_data.im <= TO_DP_DSP_DATA(yimag_stimuli_arr(index));
            wait until rising_edge(dp_clk) and valid_enable = '1';
          end loop;
        end loop;
      end if;
    end loop;

    for I in 0 to 9 loop wait until rising_edge(dp_clk); end loop;
    tb_end <= '1';
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------

  dut : entity work.iquv_iab
  generic map (
    g_sim              => true,
    g_technology       => c_tech_select_default,
    g_use_accum        => g_use_accum,
    g_nof_streams      => g_nof_streams,
    g_nof_int          => g_nof_int,
    g_in_data_w        => g_in_data_w,
    g_out_data_w       => g_out_data_w
  )
  port map (
    dp_rst         => dp_rst,
    dp_clk         => dp_clk,
    in_complex_arr => in_complex_arr,
    i_out          => i_out,
    q_out          => q_out,
    u_out          => u_out,
    v_out          => v_out
  );

  -- diag signals to make outputs visible in Modelsim wave

  diag_i_out_data <= i_out.data(c_word_w - 1 downto 0);
  diag_q_out_data <= q_out.data(c_word_w - 1 downto 0);
  diag_u_out_data <= u_out.data(c_word_w - 1 downto 0);
  diag_v_out_data <= v_out.data(c_word_w - 1 downto 0);
  diag_out_valid  <= i_out.valid;

  ------------------------------------------------------------------------------
  -- Results Checker
  ------------------------------------------------------------------------------
  p_verify : process(dp_clk)
  variable v_index : integer := 0;
  variable i_expected_iquv : std_logic_vector(c_out_iquv_w - 1 downto 0);
  variable q_expected_iquv : std_logic_vector(c_out_iquv_w - 1 downto 0);
  variable u_expected_iquv : std_logic_vector(c_out_iquv_w - 1 downto 0);
  variable v_expected_iquv : std_logic_vector(c_out_iquv_w - 1 downto 0);
  variable i_expected_iab  : std_logic_vector(c_fsd_w - 1 downto 0);
  variable q_expected_iab  : std_logic_vector(c_fsd_w - 1 downto 0);
  variable u_expected_iab  : std_logic_vector(c_fsd_w - 1 downto 0);
  variable v_expected_iab  : std_logic_vector(c_fsd_w - 1 downto 0);
  begin
    if rising_edge(dp_clk) then
      if diag_out_valid = '1' then
        activestream_dly <= activestream;  -- make sure we capture the last output valid of each cycle
        -- Integration is done in the IQUV
        if g_use_accum = false and g_nof_int = 8 then
          i_expected_iquv := u_round(TO_UVEC(i_nint8_exp_arr(v_index),c_fsd_iquv_w - 1), c_lsb_iquv_w - 1, true);
          q_expected_iquv := s_round(TO_SVEC(q_nint8_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          u_expected_iquv := s_round(TO_SVEC(u_nint8_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          v_expected_iquv := s_round(TO_SVEC(v_nint8_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
        elsif g_use_accum = false and g_nof_int = 16 then
          i_expected_iquv := u_round(TO_UVEC(i_nint16_exp_arr(v_index),c_fsd_iquv_w - 1), c_lsb_iquv_w - 1, true);
          q_expected_iquv := s_round(TO_SVEC(q_nint16_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          u_expected_iquv := s_round(TO_SVEC(u_nint16_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          v_expected_iquv := s_round(TO_SVEC(v_nint16_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
        -- Integration is done in the IAB
        elsif g_use_accum = true then
          i_expected_iquv := u_round(TO_UVEC(i_noint_exp_arr(v_index),c_fsd_iquv_w - 1), c_lsb_iquv_w - 1, true);
          q_expected_iquv := s_round(TO_SVEC(q_noint_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          u_expected_iquv := s_round(TO_SVEC(u_noint_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
          v_expected_iquv := s_round(TO_SVEC(v_noint_exp_arr(v_index),c_fsd_iquv_w), c_lsb_iquv_w, true);
        end if;
        if activestream_dly = c_allstreams then
          i_expected_iab  := TO_UVEC(to_uint(i_expected_iquv) * c_allstreams * c_nof_iab_int, c_fsd_w);
          q_expected_iab  := TO_SVEC(to_sint(q_expected_iquv) * c_allstreams * c_nof_iab_int, c_fsd_w);
          u_expected_iab  := TO_SVEC(to_sint(u_expected_iquv) * c_allstreams * c_nof_iab_int, c_fsd_w);
          v_expected_iab  := TO_SVEC(to_sint(v_expected_iquv) * c_allstreams * c_nof_iab_int, c_fsd_w);
        else
          i_expected_iab  := TO_UVEC(to_uint(i_expected_iquv) * c_nof_iab_int, c_fsd_w);
          q_expected_iab  := TO_SVEC(to_sint(q_expected_iquv) * c_nof_iab_int, c_fsd_w);
          u_expected_iab  := TO_SVEC(to_sint(u_expected_iquv) * c_nof_iab_int, c_fsd_w);
          v_expected_iab  := TO_SVEC(to_sint(v_expected_iquv) * c_nof_iab_int, c_fsd_w);
        end if;
        assert TO_UINT(i_out.data) = TO_UINT(u_round(i_expected_iab, c_lsb_w, true))
          report "Error: wrong result in I out DUT" severity ERROR;
        assert TO_SINT(q_out.data) = TO_SINT(s_round(q_expected_iab, c_lsb_w, true))
          report "Error: wrong result in Q out DUT" severity ERROR;
        assert TO_SINT(u_out.data) = TO_SINT(s_round(u_expected_iab, c_lsb_w, true))
          report "Error: wrong result in U out DUT" severity ERROR;
        assert TO_SINT(v_out.data) = TO_SINT(s_round(v_expected_iab, c_lsb_w, true))
          report "Error: wrong result in V out DUT" severity ERROR;
        --REPORT "I expected IQUV = " & integer'image(to_uint(i_expected_iquv));
        --REPORT "I expected IAB  = " & integer'image(to_uint(u_round(i_expected_iab, c_lsb_w, TRUE)));
        --REPORT "I actual        = " & integer'image(TO_UINT(i_out.data));
        --REPORT "U expected IQUV = " & integer'image(to_sint(u_expected_iquv));
        --REPORT "U expected IAB  = " & integer'image(to_sint(s_round(u_expected_iab, c_lsb_w, TRUE)));
        --REPORT "U actual        = " & integer'image(TO_sINT(u_out.data));

        --REPORT "I = " & integer'image(to_uint(i_out.data)) & ", Q = " & integer'image(to_sint(q_out.data)) &
        --     ", U = " & integer'image(to_sint(u_out.data)) & ", V = " & integer'image(to_sint(v_out.data));
        if v_index = c_nof_lines - 1 then
          v_index := 0;
        else
          v_index := v_index + 1;
        end if;
      end if;
    end if;
  end process;
end tb;
