-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
-- . Test dp_offload_tx and dp_offload_rx components
-- Description:
-- . Block generators generate data blocks that flow from dp_offload_tx to dp_offload_rx
--   instances via 10GbE (64b user interface)

library IEEE, common_lib, unb1_board_lib, dp_lib, eth_lib, tech_tse_lib, diag_lib, tr_10GbE_lib, technology_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_str_pkg.all;
use common_lib.common_mem_pkg.all;
use common_lib.common_network_layers_pkg.all;
use common_lib.common_interface_layers_pkg.all;
use unb1_board_lib.unb1_board_pkg.all;
use unb1_board_lib.unb1_board_peripherals_pkg.all;
use dp_lib.dp_stream_pkg.all;
use tech_tse_lib.tech_tse_pkg.all;
use eth_lib.eth_pkg.all;
use common_lib.common_network_layers_pkg.all;
use diag_lib.diag_pkg.all;
use common_lib.common_field_pkg.all;
use technology_lib.technology_pkg.all;

entity unb1_eth_10g is
  generic (
    g_design_name  : string  := "unb1_eth_10g";
    g_design_note  : string  := "revision info";
    g_technology   : natural := c_tech_stratixiv;
    g_sim          : boolean := false;  -- set by     ModelSim
    g_sim_unb_nr   : natural := 0;  -- set by     ModelSim
    g_sim_node_nr  : natural := 0;  -- set by     ModelSim
    g_stamp_date   : natural := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time   : natural := 0;  -- Time (HHMMSS)   -- set by QSF
    g_stamp_svn    : natural := 0  -- SVN revision    -- set by QSF
  );
  port (
    -- GENERAL
    CLK           : in    std_logic;  -- dp_clk is generated by SOPC altpll
    PPS           : in    std_logic;
    WDI           : out   std_logic;
    INTA          : inout std_logic;
    INTB          : inout std_logic;

    -- Others
    VERSION       : in    std_logic_vector(c_unb1_board_aux.version_w - 1 downto 0);
    ID            : in    std_logic_vector(c_unb1_board_aux.id_w - 1 downto 0);
    TESTIO        : inout std_logic_vector(c_unb1_board_aux.testio_w - 1 downto 0);

    -- I2C Interface to Sensors
    sens_sc       : inout std_logic;
    sens_sd       : inout std_logic;

    -- 1GbE Control Interface
    ETH_clk       : in    std_logic;
    ETH_SGIN      : in    std_logic;
    ETH_SGOUT     : out   std_logic;

    -- Transceiver clocks
    SA_CLK        : in  std_logic := '0';  -- SerDes Clock BN-BI / SI_FN

    -- Serial I/O: 10GbE receivers
    SI_FN_0_TX    : out   std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_0_RX    : in    std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_1_TX    : out   std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_1_RX    : in    std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_2_TX    : out   std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_2_RX    : in    std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_3_TX    : out   std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);
    SI_FN_3_RX    : in    std_logic_vector(c_unb1_board_ci.tr.bus_w - 1 downto 0);

    SI_FN_0_CNTRL : inout std_logic_vector(c_unb1_board_ci.tr.cntrl_w - 1 downto 0);
    SI_FN_1_CNTRL : inout std_logic_vector(c_unb1_board_ci.tr.cntrl_w - 1 downto 0);
    SI_FN_2_CNTRL : inout std_logic_vector(c_unb1_board_ci.tr.cntrl_w - 1 downto 0);
    SI_FN_3_CNTRL : inout std_logic_vector(c_unb1_board_ci.tr.cntrl_w - 1 downto 0);
    SI_FN_RSTN    : out   std_logic := '1'  -- ResetN is pulled up in the Vitesse chip, but pulled down again by external 1k resistor.
                                           -- So we need to assign a '1' to it.
  );
end unb1_eth_10g;

architecture str of unb1_eth_10g is
  -- Firmware version x.y
  constant c_fw_version                 : t_unb1_board_fw_version := (1, 0);  --

--  CONSTANT c_eth_packet_size --FIXME

  -- Revision controlled constants
  constant c_use_1GbE                   : boolean := true;
  constant c_nof_streams                : natural := 1;
  constant c_nof_10GbE_streams          : natural := c_nof_streams;
  constant c_nof_bsn_mon_streams        : natural := 5;

  constant c_data_w                     : natural := 64;  -- c_tech_tse_data_w;

  -- Block generator
  constant c_bg_addr_w                  : natural := 9;
  constant c_bg_block_size              : natural := 365;
  constant c_bg_gapsize                 : natural := c_bg_block_size / 2;  -- Full (no gaps in data) BG output rate = 200MHz * 64b = 12.8Gbps. Including gap size: (365/(365+182))*12.8Gbps=8.54Gbps.
  constant c_bg_calc_blocks_per_sync    : natural := 200000000 / (c_bg_block_size + c_bg_gapsize);
  constant c_bg_blocks_per_sync         : natural := sel_a_b(g_sim, 10, c_bg_calc_blocks_per_sync);  -- 200000*(900+100) = 200000000 cycles = 1 second
  constant c_bg_ctrl                    : t_diag_block_gen := ('0',  -- enable
                                                               '0',  -- enable_sync
                                                              TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                              TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                              TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                              TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                              TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                              TO_UVEC(                   0, c_diag_bg_bsn_init_w));

  -- dp_offload_tx
  -- . IP total length : 2948 (UDP total lenth) + 20 (Ip header length) = 2968
  -- . UDP total length: 8 (UDP header) + 20 (usr header) +  2920 (payload bytes) = 2948     -- 1488
  constant c_ip_length      : natural := c_bg_block_size * 8 + 50;  -- 2970;
  constant c_udp_length     : natural := c_bg_block_size * 8 + 30;  -- 2950;
  constant c_nof_hdr_fields : natural := 3 + 12 + 4 + 9 + 1;  -- Total header bits = 512
  constant c_hdr_field_arr  : t_common_field_arr(c_nof_hdr_fields - 1 downto 0) := ( ( field_name_pad("eth_dst_mac"        ), "  ", 48, field_default(0) ),
                                                                                   ( field_name_pad("eth_src_mac"        ), "  ", 48, field_default(0) ),
                                                                                   ( field_name_pad("eth_type"           ), "  ", 16, field_default(x"0800") ),
                                                                                   ( field_name_pad("ip_version"         ), "  ",  4, field_default(4) ),
                                                                                   ( field_name_pad("ip_header_length"   ), "  ",  4, field_default(5) ),
                                                                                   ( field_name_pad("ip_services"        ), "  ",  8, field_default(0) ),
                                                                                   ( field_name_pad("ip_total_length"    ), "  ", 16, field_default(c_ip_length) ),  -- 1508) ),
                                                                                   ( field_name_pad("ip_identification"  ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("ip_flags"           ), "  ",  3, field_default(2) ),
                                                                                   ( field_name_pad("ip_fragment_offset" ), "  ", 13, field_default(0) ),
                                                                                   ( field_name_pad("ip_time_to_live"    ), "  ",  8, field_default(127) ),
                                                                                   ( field_name_pad("ip_protocol"        ), "  ",  8, field_default(17) ),
                                                                                   ( field_name_pad("ip_header_checksum" ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("ip_src_addr"        ), "  ", 32, field_default(0) ),
                                                                                   ( field_name_pad("ip_dst_addr"        ), "  ", 32, field_default(0) ),
                                                                                   ( field_name_pad("udp_src_port"       ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("udp_dst_port"       ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("udp_total_length"   ), "  ", 16, field_default(c_udp_length) ),  -- 1488) ),
                                                                                   ( field_name_pad("udp_checksum"       ), "  ", 16, field_default(0) ),
                                                                                   ( field_name_pad("usr_sync"           ), "  ",  1, field_default(1) ),
                                                                                   ( field_name_pad("usr_bsn"            ), "  ", 60, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_0"    ), "  ",  7, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_1"    ), "  ",  9, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_2"    ), "  ", 10, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_3"    ), "  ", 33, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_4"    ), "  ",  5, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_5"    ), "  ",  8, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_field_6"    ), "  ", 27, field_default(0) ),
                                                                                   ( field_name_pad("usr_hdr_word_align" ), "  ", 16, field_default(0) ) );

  constant c_bypass_rx_filter             : boolean := true;

  constant c_hdr_field_ovr_init           : std_logic_vector(c_nof_hdr_fields - 1 downto 0) := "111" & "111111111111" & "0011" & "001111111" & "0";

  constant c_nof_crc_words                : natural := 0;
  constant c_max_nof_words_per_block      : natural := c_bg_block_size;
  constant c_min_nof_words_per_block      : natural := 1;
  constant c_def_nof_words_per_block      : natural := c_bg_block_size;
  constant c_def_nof_blocks_per_packet    : natural := 1;

  signal hdr_fields_in_arr                : t_slv_1024_arr(c_nof_streams - 1 downto 0);
  signal hdr_fields_in_rx_arr             : t_slv_1024_arr(c_nof_streams - 1 downto 0);
  signal hdr_fields_out_arr               : t_slv_1024_arr(c_nof_streams - 1 downto 0);

  -- System
  signal sa_rst                           : std_logic := '0';
  signal cs_sim                           : std_logic;
  signal xo_clk                           : std_logic;
  signal xo_rst                           : std_logic;
  signal xo_rst_n                         : std_logic;
  signal mm_clk                           : std_logic;
  signal mm_locked                        : std_logic;
  signal mm_rst                           : std_logic;
  signal dp_rst                           : std_logic;
  signal dp_clk                           : std_logic;
  signal dp_pps                           : std_logic;
  signal epcs_clk                         : std_logic;
  -- PIOs
  signal pout_wdi                         : std_logic;

  signal eth1g_tse_clk                    : std_logic;
  signal eth1g_mm_rst                     : std_logic;
  signal eth1g_tse_mosi                   : t_mem_mosi;
  signal eth1g_tse_miso                   : t_mem_miso;
  signal eth1g_reg_mosi                   : t_mem_mosi;
  signal eth1g_reg_miso                   : t_mem_miso;
  signal eth1g_reg_interrupt              : std_logic;
  signal eth1g_ram_mosi                   : t_mem_mosi;
  signal eth1g_ram_miso                   : t_mem_miso;

  signal reg_wdi_mosi                     : t_mem_mosi;
  signal reg_wdi_miso                     : t_mem_miso;
  signal reg_unb_system_info_mosi         : t_mem_mosi;
  signal reg_unb_system_info_miso         : t_mem_miso;
  signal rom_unb_system_info_mosi         : t_mem_mosi;
  signal rom_unb_system_info_miso         : t_mem_miso;
  signal reg_unb_sens_mosi                : t_mem_mosi;
  signal reg_unb_sens_miso                : t_mem_miso;

  signal reg_diag_bg_mosi                 : t_mem_mosi;
  signal reg_diag_bg_miso                 : t_mem_miso;
  signal ram_diag_bg_mosi                 : t_mem_mosi;
  signal ram_diag_bg_miso                 : t_mem_miso;

  signal reg_dp_offload_tx_mosi           : t_mem_mosi;
  signal reg_dp_offload_tx_miso           : t_mem_miso;
  signal reg_dp_offload_tx_hdr_dat_mosi   : t_mem_mosi;
  signal reg_dp_offload_tx_hdr_dat_miso   : t_mem_miso;
  signal reg_dp_offload_tx_hdr_ovr_mosi   : t_mem_mosi;
  signal reg_dp_offload_tx_hdr_ovr_miso   : t_mem_miso;
  signal reg_dp_offload_rx_hdr_dat_mosi   : t_mem_mosi;
  signal reg_dp_offload_rx_hdr_dat_miso   : t_mem_miso;

  signal reg_dp_offload_rx_filter_hdr_fields_mosi : t_mem_mosi;
  signal reg_dp_offload_rx_filter_hdr_fields_miso : t_mem_miso;

  --. 10G Receiver
  signal reg_tr_10GbE_mosi                : t_mem_mosi;
  signal reg_tr_10GbE_miso                : t_mem_miso;
  signal reg_tr_xaui_mosi                 : t_mem_mosi;
  signal reg_tr_xaui_miso                 : t_mem_miso;
  signal reg_mdio_mosi_arr                : t_mem_mosi_arr(c_unb1_board_nof_mdio - 1 downto 0);
  signal reg_mdio_miso_arr                : t_mem_miso_arr(c_unb1_board_nof_mdio - 1 downto 0);

  signal reg_mdio_0_mosi                  : t_mem_mosi;
  signal reg_mdio_0_miso                  : t_mem_miso;
  signal reg_mdio_1_mosi                  : t_mem_mosi;
  signal reg_mdio_1_miso                  : t_mem_miso;
  signal reg_mdio_2_mosi                  : t_mem_mosi;
  signal reg_mdio_2_miso                  : t_mem_miso;

  signal reg_bsn_monitor_mosi             : t_mem_mosi;
  signal reg_bsn_monitor_miso             : t_mem_miso;
  signal ram_diag_data_buffer_mosi        : t_mem_mosi;
  signal ram_diag_data_buffer_miso        : t_mem_miso;
  signal reg_diag_data_buffer_mosi        : t_mem_mosi;
  signal reg_diag_data_buffer_miso        : t_mem_miso;

  signal reg_ppsh_mosi                    : t_mem_mosi;
  signal reg_ppsh_miso                    : t_mem_miso;

  -- xonoff

  signal reg_dp_xonoff_mosi               : t_mem_mosi;
  signal reg_dp_xonoff_miso               : t_mem_miso;

  signal block_gen_src_out_arr            : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal block_gen_src_in_arr             : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);
  signal block_gen_src_in_arr_temp        : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  signal dp_offload_tx_snk_in_arr         : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_tx_snk_out_arr        : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_tx_src_out_arr        : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_tx_src_in_arr         : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal eth_checksum_10G_snk_in_arr      : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal eth_checksum_10G_snk_out_arr     : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal eth_checksum_10G_src_out_arr     : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal eth_checksum_10G_src_in_arr      : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal mms_dp_xonoff_src_out_arr        : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal mms_dp_xonoff_src_in_arr         : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_rx_restored_src_out_arr : t_dp_sosi_arr(c_nof_10GbE_streams - 1 downto 0) := (others => c_dp_sosi_rst);

  signal dp_offload_rx_snk_in_arr         : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_rx_snk_out_arr        : t_dp_siso_arr(c_nof_streams - 1 downto 0);

  signal dp_offload_rx_src_out_arr        : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_rx_src_in_arr         : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  signal dp_offload_rx_filter_src_out_arr : t_dp_sosi_arr(c_nof_streams - 1 downto 0);
  signal dp_offload_rx_filter_src_in_arr  : t_dp_siso_arr(c_nof_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  signal bsn_monitor_snk_in_arr           : t_dp_sosi_arr(c_nof_bsn_mon_streams - 1 downto 0);
  signal bsn_monitor_snk_out_arr          : t_dp_siso_arr(c_nof_bsn_mon_streams - 1 downto 0) := (others => c_dp_siso_rdy);

  -- Interface: 10GbE
  signal xaui_tx_arr                      : t_xaui_arr(c_nof_10GbE_streams - 1 downto 0);
  signal xaui_rx_arr                      : t_xaui_arr(c_nof_10GbE_streams - 1 downto 0);
  signal unb_xaui_tx_arr                  : t_unb1_board_xaui_sl_2arr(c_nof_10GbE_streams - 1 downto 0);
  signal unb_xaui_rx_arr                  : t_unb1_board_xaui_sl_2arr(c_nof_10GbE_streams - 1 downto 0);
  signal mdio_mdc_arr                     : std_logic_vector(c_nof_10GbE_streams - 1 downto 0);
  signal mdio_mdat_in_arr                 : std_logic_vector(c_nof_10GbE_streams - 1 downto 0);
  signal mdio_mdat_oen_arr                : std_logic_vector(c_nof_10GbE_streams - 1 downto 0);

  signal TEST_STOP                        : std_logic_vector(2 downto 0 );
  signal TEST_ERROR                       : std_logic_vector(2 downto 0 );
  signal TEST_FIFO_FULL                   : std_logic_vector(1 downto 0 );
  signal TEST_BLOCK_RD                    : std_logic_vector(2 downto 0 );
  signal address                          : std_logic_vector(18 downto 0 );
  signal read_data                        : std_logic_vector(31 downto 0 );
  signal read_en                          : std_logic;
  signal write_en                         : std_logic;
  signal write_data                       : std_logic_vector(31 downto 0 );

  signal fi_snk_out                       : t_dp_siso;
  signal fi_snk_in                        : t_dp_sosi;
  signal fi_src_in                        : t_dp_siso;
  signal fi_src_out                       : t_dp_sosi;
  signal fo_snk_out                       : t_dp_siso;
  signal fo_snk_in                        : t_dp_sosi;
  signal fo_src_in                        : t_dp_siso;
  signal fo_src_out                       : t_dp_sosi;
begin
    -----------------------------------------------------------------------------
    -- Interface : 10GbE
    -----------------------------------------------------------------------------
    -- Wire together different types
    gen_wires: for i in 0 to c_nof_10GbE_streams - 1 generate
      unb_xaui_tx_arr(i) <= xaui_tx_arr(i);
      xaui_rx_arr(i)     <= unb_xaui_rx_arr(i);
    end generate;

    u_front_io : entity unb1_board_lib.unb1_board_front_io
    generic map (
      g_nof_xaui => c_nof_10GbE_streams
    )
    port map (
      xaui_tx_arr       => unb_xaui_tx_arr,
      xaui_rx_arr       => unb_xaui_rx_arr,

      mdio_mdc_arr      => mdio_mdc_arr,
      mdio_mdat_in_arr  => mdio_mdat_in_arr,
      mdio_mdat_oen_arr => mdio_mdat_oen_arr,

      -- Serial I/O
      SI_FN_0_TX        => SI_FN_0_TX,
      SI_FN_0_RX        => SI_FN_0_RX,
      SI_FN_1_TX        => SI_FN_1_TX,
      SI_FN_1_RX        => SI_FN_1_RX,
      SI_FN_2_TX        => SI_FN_2_TX,
      SI_FN_2_RX        => SI_FN_2_RX,

      SI_FN_0_CNTRL     => SI_FN_0_CNTRL,
      SI_FN_1_CNTRL     => SI_FN_1_CNTRL,
      SI_FN_2_CNTRL     => SI_FN_2_CNTRL,
      SI_FN_3_CNTRL     => SI_FN_3_CNTRL
    );

    u_areset_sa_rst : entity common_lib.common_areset
    generic map(
      g_rst_level => '1',
      g_delay_len => 4
    )
    port map(
      clk     => SA_CLK,
      in_rst  => '0',
      out_rst => sa_rst
    );

    u_tr_10GbE: entity tr_10GbE_lib.tr_10GbE
    generic map(
      g_sim             => g_sim,
      g_sim_level       => 1,
      g_nof_macs        => c_nof_10GbE_streams,
      g_use_mdio        => true
    )
    port map (
      -- Transceiver PLL reference clock
      tr_ref_clk_156    => SA_CLK,
      tr_ref_rst_156    => sa_rst,

      -- Calibration & reconfig clock
      cal_rec_clk       => mm_clk,

      -- MM interface
      mm_rst            => mm_rst,
      mm_clk            => mm_clk,

      reg_mac_mosi      => reg_tr_10GbE_mosi,
      reg_mac_miso      => reg_tr_10GbE_miso,

      xaui_mosi         => reg_tr_xaui_mosi,
      xaui_miso         => reg_tr_xaui_miso,

      mdio_mosi_arr     => reg_mdio_mosi_arr(c_nof_10GbE_streams - 1 downto 0),
      mdio_miso_arr     => reg_mdio_miso_arr(c_nof_10GbE_streams - 1 downto 0),

      -- DP interface
      dp_rst            => dp_rst,
      dp_clk            => dp_clk,

      -- Data received by 10G
      src_out_arr       => dp_offload_rx_snk_in_arr,
      src_in_arr        => dp_offload_rx_snk_out_arr,

      -- Data to be send by 10G
      snk_out_arr       =>  mms_dp_xonoff_src_in_arr,  -- dp_offload_tx_src_in_arr,
      snk_in_arr        =>  mms_dp_xonoff_src_out_arr,  -- dp_offload_tx_src_out_arr,

      -- Serial XAUI IO
      xaui_tx_arr       => xaui_tx_arr,
      xaui_rx_arr       => xaui_rx_arr,

      -- MDIO interface
      mdio_rst          => SI_FN_RSTN,
      mdio_mdc_arr      => mdio_mdc_arr,
      mdio_mdat_in_arr  => mdio_mdat_in_arr,
      mdio_mdat_oen_arr => mdio_mdat_oen_arr
    );

  -----------------------------------------------------------------------------
  -- TX: Block generator
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : entity diag_lib.mms_diag_block_gen
  generic map (
    g_nof_streams        => c_nof_streams,
    g_buf_dat_w          => c_data_w,
    g_buf_addr_w         => c_bg_addr_w,  -- ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
    g_file_name_prefix   => "UNUSED",
    g_diag_block_gen_rst => c_bg_ctrl
  )
  port map (
    mm_rst           => mm_rst,
    mm_clk           => mm_clk,

    dp_rst           => dp_rst,
    dp_clk           => dp_clk,

    out_sosi_arr     => block_gen_src_out_arr,
    out_siso_arr     => block_gen_src_in_arr,

    reg_bg_ctrl_mosi => reg_diag_bg_mosi,
    reg_bg_ctrl_miso => reg_diag_bg_miso,
    ram_bg_data_mosi => ram_diag_bg_mosi,
    ram_bg_data_miso => ram_diag_bg_miso
  );

  gen_dp_fifo_sc : for i in 0 to c_nof_streams - 1 generate
    u_dp_fifo_sc : entity dp_lib.dp_fifo_sc
    generic map (
      g_bsn_w       => c_dp_stream_bsn_w,
      g_use_bsn     => true,
      g_use_empty   => true,
      g_use_channel => true,
      g_use_error   => true,
      g_use_sync    => true,
      g_data_w      => c_data_w,
      g_fifo_size   => 3 * c_bg_block_size
    )
    port map (
      rst         => dp_rst,
      clk         => dp_clk,
      -- ST sink (BG)
      snk_out     => block_gen_src_in_arr(i),
      snk_in      => block_gen_src_out_arr(i),
      -- ST source (tx_offload)
      src_in      => dp_offload_tx_snk_out_arr(i),
      src_out     => dp_offload_tx_snk_in_arr(i)
    );
  end generate;

  -----------------------------------------------------------------------------
  -- TX: dp_offload_tx
  -----------------------------------------------------------------------------
  u_dp_offload_tx : entity dp_lib.dp_offload_tx
  generic map (
    g_nof_streams               => c_nof_10GbE_streams,
    g_data_w                    => c_data_w,
    g_use_complex               => false,
--    g_max_nof_words_per_block   => c_max_nof_words_per_block,
    g_nof_words_per_block       => c_def_nof_words_per_block,
--    g_max_nof_blocks_per_packet => c_max_nof_blocks_per_packet,
    g_nof_blocks_per_packet     => c_def_nof_blocks_per_packet,
    g_hdr_field_arr             => c_hdr_field_arr,
    g_hdr_field_sel             => c_hdr_field_ovr_init,
    g_use_post_split_fifo       => false
   )
  port map (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,

    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

--    reg_mosi              => reg_dp_offload_tx_mosi,
--    reg_miso              => reg_dp_offload_tx_miso,

    reg_hdr_dat_mosi      => reg_dp_offload_tx_hdr_dat_mosi,
    reg_hdr_dat_miso      => reg_dp_offload_tx_hdr_dat_miso,

    snk_in_arr            => dp_offload_tx_snk_in_arr,
    snk_out_arr           => dp_offload_tx_snk_out_arr,

    src_out_arr           => dp_offload_tx_src_out_arr,
    src_in_arr            => dp_offload_tx_src_in_arr,

    hdr_fields_in_arr     => hdr_fields_in_arr
  );

  dp_offload_tx_src_in_arr <= eth_checksum_10G_snk_out_arr;
  eth_checksum_10G_snk_in_arr <= dp_offload_tx_src_out_arr;

  gen_insert_checksum : for i in 0 to c_nof_streams - 1 generate
    u_insert_checksum : entity eth_lib.eth_checksum_10G
    port map (
      rst      => dp_rst,
      clk      => dp_clk,

      src_out  => eth_checksum_10G_src_out_arr(i),
      src_in   => eth_checksum_10G_src_in_arr(i),

      snk_in   => eth_checksum_10G_snk_in_arr(i),
      snk_out  => eth_checksum_10G_snk_out_arr(i)
    );
  end generate;

  u_mms_dp_xonoff : entity dp_lib.mms_dp_xonoff
  generic map(
    g_nof_streams     => 1,
    g_combine_streams => true,
    g_bypass          => false,
    g_timeout_time    => 10,
    g_sim             => g_sim
  )
  port map(
    mm_rst       => mm_rst,
    mm_clk       => mm_clk,

    reg_mosi     => reg_dp_xonoff_mosi,
    reg_miso     => reg_dp_xonoff_miso,

    dp_rst       => dp_rst,
    dp_clk       => dp_clk,

    snk_out_arr  => eth_checksum_10G_src_in_arr,
    snk_in_arr   => eth_checksum_10G_src_out_arr,

    src_in_arr   => mms_dp_xonoff_src_in_arr,
    src_out_arr  => mms_dp_xonoff_src_out_arr
  );

  gen_hdr_in_fields : for i in 0 to c_nof_streams - 1 generate
    -- dst = src
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "eth_src_mac"       ) downto field_lo(c_hdr_field_arr, "eth_src_mac"        )) <= x"00228608" & B"000" & ID(7 downto 3) & RESIZE_UVEC(ID(2 downto 0), c_byte_w);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "eth_dst_mac"       ) downto field_lo(c_hdr_field_arr, "eth_dst_mac"        )) <= x"00228608" & B"000" & ID(7 downto 3) & RESIZE_UVEC(ID(2 downto 0), c_byte_w);

    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "ip_src_addr"       ) downto field_lo(c_hdr_field_arr, "ip_src_addr"        )) <= x"0A63" & B"000" & ID(7 downto 3) & INCR_UVEC(RESIZE_UVEC(ID(2 downto 0), c_byte_w), 1);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "ip_dst_addr"       ) downto field_lo(c_hdr_field_arr, "ip_dst_addr"        )) <= x"0A63" & B"000" & ID(7 downto 3) & INCR_UVEC(RESIZE_UVEC(ID(2 downto 0), c_byte_w), 1);

    -- dst port goes through 4000,4001,4002
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "udp_src_port"      ) downto field_lo(c_hdr_field_arr, "udp_src_port"       )) <= TO_UVEC(4000 + i, 16);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "udp_dst_port"      ) downto field_lo(c_hdr_field_arr, "udp_dst_port"       )) <= TO_UVEC(4000 + i, 16);

    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "usr_sync"          ) downto field_lo(c_hdr_field_arr, "usr_sync"           )) <= slv(dp_offload_tx_snk_in_arr(i).sync);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "usr_bsn"           ) downto field_lo(c_hdr_field_arr, "usr_bsn"            )) <= dp_offload_tx_snk_in_arr(i).bsn(59 downto 0);

    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "usr_hdr_word_align") downto field_lo(c_hdr_field_arr, "usr_hdr_word_align" )) <= TO_UVEC(0, 16);
    hdr_fields_in_arr(i)(field_hi(c_hdr_field_arr, "ip_total_length"   ) downto field_lo(c_hdr_field_arr, "ip_total_length"    )) <= TO_UVEC(c_ip_length, 16);
  end generate;

  -----------------------------------------------------------------------------
  -- RX: dp_offload_rx
  -----------------------------------------------------------------------------
  u_dp_offload_rx : entity dp_lib.dp_offload_rx
  generic map (
    g_nof_streams         => c_nof_streams,
    g_data_w              => c_data_w,
    g_hdr_field_arr       => c_hdr_field_arr,
    g_remove_crc          => c_use_1GbE,
    g_crc_nof_words       => c_nof_crc_words
   )
  port map (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,

    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

    reg_hdr_dat_mosi      => reg_dp_offload_rx_hdr_dat_mosi,
    reg_hdr_dat_miso      => reg_dp_offload_rx_hdr_dat_miso,

    snk_in_arr            => dp_offload_rx_snk_in_arr,
    snk_out_arr           => dp_offload_rx_snk_out_arr,

    src_out_arr           => dp_offload_rx_src_out_arr,
    src_in_arr            => dp_offload_rx_src_in_arr,

    hdr_fields_out_arr    => hdr_fields_out_arr
  );

  -- Restore the sync and bsn signals to the offload_rx output
  gen_restore_sync_bsn : for i in 0 to c_nof_10GbE_streams - 1 generate
    dp_offload_rx_restored_src_out_arr(i).sync <=          sl(hdr_fields_out_arr(i)(field_hi(c_hdr_field_arr, "usr_sync") downto field_lo(c_hdr_field_arr, "usr_sync" )));
    dp_offload_rx_restored_src_out_arr(i).bsn  <= RESIZE_UVEC(hdr_fields_out_arr(i)(field_hi(c_hdr_field_arr, "usr_bsn" ) downto field_lo(c_hdr_field_arr, "usr_bsn"  )), c_dp_stream_bsn_w);

    dp_offload_rx_restored_src_out_arr(i).data  <= dp_offload_rx_src_out_arr(i).data;
    dp_offload_rx_restored_src_out_arr(i).valid <= dp_offload_rx_src_out_arr(i).valid;
    dp_offload_rx_restored_src_out_arr(i).sop   <= dp_offload_rx_src_out_arr(i).sop;
    dp_offload_rx_restored_src_out_arr(i).eop   <= dp_offload_rx_src_out_arr(i).eop;
    dp_offload_rx_restored_src_out_arr(i).err   <= dp_offload_rx_src_out_arr(i).err;
  end generate;

  -------------------------
  --WORK IN PROGRESS
  -------------------------
  process(hdr_fields_in_arr)
  begin
    hdr_fields_in_rx_arr <= hdr_fields_in_arr;
    hdr_fields_in_rx_arr(0)(field_hi(c_hdr_field_arr, "eth_dst_mac" ) downto field_lo(c_hdr_field_arr, "eth_dst_mac" )) <= x"00074306C700";  -- jop63
    hdr_fields_in_rx_arr(0)(field_hi(c_hdr_field_arr, "ip_dst_addr" ) downto field_lo(c_hdr_field_arr, "ip_dst_addr" )) <= x"c0a80102";  -- jop63
  end process;

  -----------------------------------------------------------------------------
  -- RX: dp_offload_rx_filter
  -----------------------------------------------------------------------------
  u_header_check : entity dp_lib.dp_offload_rx_filter_mm
  generic map(
    g_bypass							 => c_bypass_rx_filter,
    g_nof_streams         => c_nof_streams,  -- : POSITIVE;
    g_data_w              => c_data_w,  -- : NATURAL;
    g_hdr_field_arr       => c_hdr_field_arr  -- : t_common_field_arr
  )
  port map(

    dp_rst             => dp_rst,
    dp_clk             => dp_clk,

    mm_rst             => mm_rst,
    mm_clk             => mm_clk,

    reg_dp_offload_rx_filter_hdr_fields_mosi => reg_dp_offload_rx_filter_hdr_fields_mosi,
    reg_dp_offload_rx_filter_hdr_fields_miso => reg_dp_offload_rx_filter_hdr_fields_miso,

    snk_in_arr              => dp_offload_rx_restored_src_out_arr,
    snk_out_arr             => dp_offload_rx_src_in_arr,

    src_out_arr             => dp_offload_rx_filter_src_out_arr,
    src_in_arr              => dp_offload_rx_filter_src_in_arr,

    hdr_fields_to_check_arr => hdr_fields_out_arr,

    hdr_fields_val          => '1'
  );

  u_dp_bsn_monitor : entity dp_lib.mms_dp_bsn_monitor
  generic map (
    g_nof_streams        => c_nof_bsn_mon_streams,
    g_cross_clock_domain => true,
    g_sync_timeout       => c_bg_blocks_per_sync * (c_bg_block_size + c_bg_gapsize),
    g_cnt_sop_w          => ceil_log2(c_bg_blocks_per_sync + 1),
    g_cnt_valid_w        => ceil_log2(c_bg_blocks_per_sync * c_bg_block_size+1),
    g_log_first_bsn      => true
  )
  port map (
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_mosi,
    reg_miso    => reg_bsn_monitor_miso,

    dp_rst      => dp_rst,
    dp_clk      => dp_clk,
    in_siso_arr => bsn_monitor_snk_out_arr,
    in_sosi_arr => bsn_monitor_snk_in_arr
  );

  bsn_monitor_snk_in_arr(0)  <= dp_offload_tx_snk_in_arr(0);
  bsn_monitor_snk_out_arr(0) <= dp_offload_tx_snk_out_arr(0);

  bsn_monitor_snk_in_arr(1)  <= dp_offload_tx_src_out_arr(0);
  bsn_monitor_snk_out_arr(1) <= dp_offload_tx_src_in_arr(0);

  bsn_monitor_snk_in_arr(2)  <= dp_offload_rx_snk_in_arr(0);
  bsn_monitor_snk_out_arr(2) <= dp_offload_rx_snk_out_arr(0);

  bsn_monitor_snk_in_arr(3)  <= dp_offload_rx_restored_src_out_arr(0);
  bsn_monitor_snk_out_arr(3) <= dp_offload_rx_src_in_arr(0);

  bsn_monitor_snk_in_arr(4)  <= block_gen_src_out_arr(0);
  bsn_monitor_snk_out_arr(4) <= block_gen_src_in_arr(0);

  --dp_offload_rx_src_in_arr <= (OTHERS=>c_dp_siso_rdy);
  dp_offload_rx_filter_src_in_arr <= (others => c_dp_siso_rdy);

  u_diag_data_buffer : entity diag_lib.mms_diag_data_buffer
  generic map (
    g_nof_streams  => c_nof_streams,
    g_data_w       => c_data_w,
    g_buf_nof_data => 1024,
    g_buf_use_sync => false
  )
  port map (
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,

    ram_data_buf_mosi => ram_diag_data_buffer_mosi,
    ram_data_buf_miso => ram_diag_data_buffer_miso,
    reg_data_buf_mosi => reg_diag_data_buffer_mosi,
    reg_data_buf_miso => reg_diag_data_buffer_miso,

    in_sync           => dp_offload_rx_filter_src_out_arr(0).sync,  -- dp_offload_rx_src_out_arr(0).sync,
    in_sosi_arr       => dp_offload_rx_filter_src_out_arr  -- dp_offload_rx_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl_unb1_board : entity unb1_board_lib.ctrl_unb1_board
  generic map (
    g_sim                     => g_sim,
    g_sim_flash_model         => false,
    g_design_name             => g_design_name,
    g_design_note             => g_design_note,
    g_fw_version              => c_fw_version,
    g_stamp_date              => g_stamp_date,
    g_stamp_time              => g_stamp_time,
    g_stamp_svn               => g_stamp_svn,
    g_mm_clk_freq             => c_unb1_board_mm_clk_freq_50M,
    g_udp_offload             => false,
    g_udp_offload_nof_streams => c_nof_streams,
    g_dp_clk_use_pll          => true,
    g_xo_clk_use_pll          => true
  )
  port map (
   -- Clock and reset signals

    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk_out               => mm_clk,
    mm_clk                   => mm_clk,
    mm_rst                   => mm_rst,

    mm_locked                => mm_locked,
    mm_locked_out            => mm_locked,

    epcs_clk                 => epcs_clk,
    epcs_clk_out             => epcs_clk,

    dp_rst                   => dp_rst,
    dp_clk                   => dp_clk,
    dp_pps                   => dp_pps,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,

    -- PIOs
    pout_wdi                 => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,

    -- eth1g
    eth1g_tse_clk_out        => eth1g_tse_clk,
    eth1g_tse_clk            => eth1g_tse_clk,
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso,
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso,

    -- UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,

    -- UniBoard FPGA pins
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
 u_inst_mmm_unb1_eth_10g : entity work.mmm_unb1_eth_10g
   generic map(
     g_sim         => g_sim,
     g_sim_unb_nr  => g_sim_unb_nr,
     g_sim_node_nr => g_sim_node_nr
   )
   port map(
     mm_clk                                   =>  mm_clk,
     mm_rst                                   =>  mm_rst,
     pout_wdi                                 =>  pout_wdi,
     reg_wdi_mosi                             =>  reg_wdi_mosi,
     reg_wdi_miso                             =>  reg_wdi_miso,
     reg_unb_system_info_mosi                 =>  reg_unb_system_info_mosi,
     reg_unb_system_info_miso                 =>  reg_unb_system_info_miso,
     rom_unb_system_info_mosi                 =>  rom_unb_system_info_mosi,
     rom_unb_system_info_miso                 =>  rom_unb_system_info_miso,
     reg_unb_sens_mosi                        =>  reg_unb_sens_mosi,
     reg_unb_sens_miso                        =>  reg_unb_sens_miso,
     reg_ppsh_mosi                            =>  reg_ppsh_mosi,
     reg_ppsh_miso                            =>  reg_ppsh_miso,
     eth1g_mm_rst                             =>  eth1g_mm_rst,
     eth1g_reg_interrupt                      =>  eth1g_reg_interrupt,
     eth1g_ram_mosi                           =>  eth1g_ram_mosi,
     eth1g_ram_miso                           =>  eth1g_ram_miso,
     eth1g_reg_mosi                           =>  eth1g_reg_mosi,
     eth1g_reg_miso                           =>  eth1g_reg_miso,
     eth1g_tse_mosi                           =>  eth1g_tse_mosi,
     eth1g_tse_miso                           =>  eth1g_tse_miso,
     reg_bsn_monitor_mosi                     =>  reg_bsn_monitor_mosi,
     reg_bsn_monitor_miso                     =>  reg_bsn_monitor_miso,
     reg_dp_offload_tx_mosi                   =>  reg_dp_offload_tx_mosi,
     reg_dp_offload_tx_miso                   =>  reg_dp_offload_tx_miso,
     reg_dp_offload_tx_hdr_dat_mosi           =>  reg_dp_offload_tx_hdr_dat_mosi,
     reg_dp_offload_tx_hdr_dat_miso           =>  reg_dp_offload_tx_hdr_dat_miso,
     reg_dp_offload_tx_hdr_ovr_mosi           =>  reg_dp_offload_tx_hdr_ovr_mosi,
     reg_dp_offload_tx_hdr_ovr_miso           =>  reg_dp_offload_tx_hdr_ovr_miso,
     reg_dp_offload_rx_hdr_dat_mosi           =>  reg_dp_offload_rx_hdr_dat_mosi,
     reg_dp_offload_rx_hdr_dat_miso           =>  reg_dp_offload_rx_hdr_dat_miso,
     reg_dp_offload_rx_filter_hdr_fields_mosi =>  reg_dp_offload_rx_filter_hdr_fields_mosi,
     reg_dp_offload_rx_filter_hdr_fields_miso =>  reg_dp_offload_rx_filter_hdr_fields_miso,
     reg_diag_data_buffer_mosi                =>  reg_diag_data_buffer_mosi,
     reg_diag_data_buffer_miso                =>  reg_diag_data_buffer_miso,
     ram_diag_data_buffer_mosi                =>  ram_diag_data_buffer_mosi,
     ram_diag_data_buffer_miso                =>  ram_diag_data_buffer_miso,
     reg_diag_bg_mosi                         =>  reg_diag_bg_mosi,
     reg_diag_bg_miso                         =>  reg_diag_bg_miso,
     ram_diag_bg_mosi                         =>  ram_diag_bg_mosi,
     ram_diag_bg_miso                         =>  ram_diag_bg_miso,
     reg_mdio_0_mosi                          =>  reg_mdio_0_mosi,
     reg_mdio_0_miso                          =>  reg_mdio_0_miso,
     reg_mdio_1_mosi                          =>  reg_mdio_1_mosi,
     reg_mdio_1_miso                          =>  reg_mdio_1_miso,
     reg_mdio_2_mosi                          =>  reg_mdio_2_mosi,
     reg_mdio_2_miso                          =>  reg_mdio_2_miso,
     reg_tr_10gbe_mosi                        =>  reg_tr_10gbe_mosi,
     reg_tr_10gbe_miso                        =>  reg_tr_10gbe_miso,
     reg_tr_xaui_mosi                         =>  reg_tr_xaui_mosi,
     reg_tr_xaui_miso                         =>  reg_tr_xaui_miso,
     reg_dp_xonoff_mosi                       =>  reg_dp_xonoff_mosi,
     reg_dp_xonoff_miso                       =>  reg_dp_xonoff_miso
   );

  reg_mdio_mosi_arr(0) <= reg_mdio_0_mosi;
  reg_mdio_mosi_arr(1) <= reg_mdio_1_mosi;
  reg_mdio_mosi_arr(2) <= reg_mdio_2_mosi;

  reg_mdio_0_miso <= reg_mdio_miso_arr(0);
  reg_mdio_1_miso <= reg_mdio_miso_arr(1);
  reg_mdio_2_miso <= reg_mdio_miso_arr(2);
end str;
