-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

library IEEE, common_lib, dp_lib, technology_lib, tech_tse_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;
use common_lib.common_mem_pkg.all;
use common_lib.tb_common_mem_pkg.all;
use common_lib.common_str_pkg.all;
use common_lib.common_lfsr_sequences_pkg.all;
use common_lib.common_network_layers_pkg.all;
use common_lib.common_network_total_header_pkg.all;
use common_lib.tb_common_pkg.all;
use dp_lib.dp_stream_pkg.all;
use dp_lib.tb_dp_pkg.all;
use technology_lib.technology_select_pkg.all;
use tech_tse_lib.tech_tse_pkg.all;
use tech_tse_lib.tb_tech_tse_pkg.all;
use WORK.eth_pkg.all;

entity tb_eth_udp_offload is
  generic (
    g_data_w       : natural := c_word_w;
    g_symbol_w     : natural := 8;
    g_in_en        : t_dp_flow_control_enum := e_random
  );
end tb_eth_udp_offload;

architecture tb of tb_eth_udp_offload is
  -- tb default
  constant c_rl               : natural := 1;
  constant c_pulse_active     : natural := 1;
  constant c_pulse_period     : natural := 7;
  constant c_technology_dut   : natural := c_tech_select_default;

  -- tb specific
  constant c_nof_repeat       : natural := 10;
  constant c_bsn_w            : natural := 16;
  constant c_symbol_init      : natural := 0;
  constant c_symbol_mod       : integer := 2**g_symbol_w;  -- used to avoid TO_UVEC warning for smaller g_symbol_w : "NUMERIC_STD.TO_UNSIGNED: vector truncated"
  constant c_err_init         : natural := 0;
  constant c_sync_period      : natural := 7;
  constant c_sync_offset      : natural := 2;

  -- clock and reset
  constant c_mm_clk_period    : time := 20 ns;
  constant c_st_clk_period    : time := 5 ns;
  constant c_eth_clk_period   : time := 8 ns;  -- 125 MHz

  -- ETH CONSTANTS
  -- ===========================================================================================================================================================

  -- Payload user data
  constant c_tb_nof_data        : natural := 1440;  -- nof UDP user data, nof ping padding data. NOTE: non-multiples of g_data_w/g_symbol_w not supported as dp_packet_enc/dec do not support encoding/decoding empty
  constant c_tb_ip_nof_data     : natural := c_network_udp_header_len + c_tb_nof_data;  -- nof IP data,

  -- Headers
  -- . Ethernet header
  constant c_lcu_src_mac        : std_logic_vector(c_network_eth_mac_slv'range) := X"10FA01020300";
  constant c_dut_src_mac        : std_logic_vector(c_network_eth_mac_slv'range) := X"123456789ABC";  -- = 12-34-56-78-9A-BC
  constant c_dut_src_mac_hi     : natural := TO_UINT(c_dut_src_mac(c_network_eth_mac_addr_w - 1 downto c_word_w));
  constant c_dut_src_mac_lo     : natural := TO_UINT(c_dut_src_mac(                c_word_w - 1 downto        0));

  constant c_tx_eth_header      : t_network_eth_header := (dst_mac    => c_dut_src_mac,
                                                           src_mac    => c_lcu_src_mac,
                                                           eth_type   => TO_UVEC(c_network_eth_type_ip, c_network_eth_type_w));  -- TO_UVEC(c_dut_ethertype, c_network_eth_type_w));
  -- . IP header
  constant c_lcu_ip_addr        : natural := 16#05060708#;  -- = 05:06:07:08
  constant c_dut_ip_addr        : natural := 16#01020304#;
  constant c_tb_ip_total_length : natural := c_network_ip_total_length + c_tb_ip_nof_data;
  constant c_tb_ip_protocol     : natural := c_network_ip_protocol_udp;  -- sel_a_b(g_data_type-c_tb_tech_tse_data_type_ping, c_network_ip_protocol_udp, c_network_ip_protocol_icmp);  -- support only ping protocol or UDP protocol over IP

  constant c_tx_ip_header       : t_network_ip_header := (version         => TO_UVEC(c_network_ip_version,         c_network_ip_version_w),
                                                          header_length   => TO_UVEC(c_network_ip_header_length,   c_network_ip_header_length_w),
                                                          services        => TO_UVEC(c_network_ip_services,        c_network_ip_services_w),
                                                          total_length    => TO_UVEC(c_tb_ip_total_length,         c_network_ip_total_length_w),
                                                          identification  => TO_UVEC(c_network_ip_identification,  c_network_ip_identification_w),
                                                          flags           => TO_UVEC(c_network_ip_flags,           c_network_ip_flags_w),
                                                          fragment_offset => TO_UVEC(c_network_ip_fragment_offset, c_network_ip_fragment_offset_w),
                                                          time_to_live    => TO_UVEC(c_network_ip_time_to_live,    c_network_ip_time_to_live_w),
                                                          protocol        => TO_UVEC(c_tb_ip_protocol,             c_network_ip_protocol_w),
                                                          header_checksum => TO_UVEC(c_network_ip_header_checksum, c_network_ip_header_checksum_w),  -- init value (or try 0xEBBD = 60349)
                                                          src_ip_addr     => TO_UVEC(c_lcu_ip_addr,                c_network_ip_addr_w),
                                                          dst_ip_addr     => TO_UVEC(c_dut_ip_addr,                c_network_ip_addr_w));

  -- . UDP header
  constant c_dut_udp_port_ctrl   : natural := 11;  -- ETH demux UDP for control
  constant c_dut_udp_port_st0    : natural := 57;  -- ETH demux UDP port 0
  constant c_dut_udp_port_st1    : natural := 58;  -- ETH demux UDP port 1
  constant c_dut_udp_port_st2    : natural := 59;  -- ETH demux UDP port 2
  constant c_dut_udp_port_en     : natural := 16#10000#;  -- ETH demux UDP port enable bit 16
  constant c_lcu_udp_port        : natural := 10;  -- UDP port used for src_port
  constant c_dut_udp_port_st     : natural := c_dut_udp_port_st0;  -- UDP port used for dst_port
  constant c_tb_udp_total_length : natural := c_network_udp_total_length + c_tb_nof_data;
  constant c_tx_udp_header       : t_network_udp_header := (src_port     => TO_UVEC(c_lcu_udp_port,         c_network_udp_port_w),
                                                            dst_port     => TO_UVEC(c_dut_udp_port_st,      c_network_udp_port_w),  -- or use c_dut_udp_port_ctrl
                                                            total_length => TO_UVEC(c_tb_udp_total_length,  c_network_udp_total_length_w),
                                                            checksum     => TO_UVEC(c_network_udp_checksum, c_network_udp_checksum_w));  -- init value

  constant c_word_align          : std_logic_vector(c_network_total_header_32b_align_w - 1 downto 0) := (others => '0');
  constant c_total_hdr_slv       : std_logic_vector(c_network_total_header_32b_nof_words * c_word_w - 1 downto 0) := c_word_align                   &
                                                                                                                 c_tx_eth_header.dst_mac        &
                                                                                                                 c_tx_eth_header.src_mac        &
                                                                                                                 c_tx_eth_header.eth_type       &
                                                                                                                 c_tx_ip_header.version         &
                                                                                                                 c_tx_ip_header.header_length   &
                                                                                                                 c_tx_ip_header.services        &
                                                                                                                 c_tx_ip_header.total_length    &
                                                                                                                 c_tx_ip_header.identification  &
                                                                                                                 c_tx_ip_header.flags           &
                                                                                                                 c_tx_ip_header.fragment_offset &
                                                                                                                 c_tx_ip_header.time_to_live    &
                                                                                                                 c_tx_ip_header.protocol        &
                                                                                                                 c_tx_ip_header.header_checksum &
                                                                                                                 c_tx_ip_header.src_ip_addr     &
                                                                                                                 c_tx_ip_header.dst_ip_addr     &
                                                                                                                 c_tx_udp_header.src_port       &
                                                                                                                 c_tx_udp_header.dst_port       &
                                                                                                                 c_tx_udp_header.total_length   &
                                                                                                                 c_tx_udp_header.checksum;

  -- ===========================================================================================================================================================

  -- TSE constants
  constant c_promis_en          : boolean := false;
  constant c_tx_ready_latency   : natural := c_tech_tse_tx_ready_latency;  -- 0, 1 are supported, must match TSE MAC c_tech_tse_tx_ready_latency

  -- ETH control
  constant c_dut_control_rx_en   : natural := 2**c_eth_mm_reg_control_bi.rx_en;

  -- ETH TSE interface
  signal eth_psc_access      : std_logic;

  signal tb_eth_hdr           : t_network_eth_header := c_tx_eth_header;
  signal tb_ip_hdr            : t_network_ip_header  := c_tx_ip_header;
  signal tb_udp_hdr           : t_network_udp_header := c_tx_udp_header;

  signal eth_clk              : std_logic := '0';  -- tse reference clock
  signal mm_clk               : std_logic := '0';
  signal mm_rst               : std_logic;
  signal st_rst               : std_logic;
  signal st_clk               : std_logic := '0';

  signal tb_end               : std_logic := '0';

  signal random_0             : std_logic_vector(14 downto 0) := (others => '0');  -- use different lengths to have different random sequences
  signal random_1             : std_logic_vector(15 downto 0) := (others => '0');  -- use different lengths to have different random sequences
  signal pulse_0              : std_logic;
  signal pulse_1              : std_logic;
  signal pulse_en             : std_logic := '1';

  signal in_en                : std_logic := '1';

  -- tb verify
  signal verify_en            : std_logic := '0';
  signal verify_done          : std_logic := '0';

  signal prev_udp_rx_ready    : std_logic_vector(0 to c_rl);
  signal prev_udp_rx_data     : std_logic_vector(g_data_w - 1 downto 0);

  signal out_gap              : std_logic := '1';

  -- dp_hdr_insert/remove signals
  signal reg_hdr_mosi         : t_mem_mosi;
  signal ram_hdr_mosi         : t_mem_mosi;
  signal ram_hdr_miso         : t_mem_miso;

  -- ETH TSE interface
  signal eth_tse_miso        : t_mem_miso;
  signal eth_tse_mosi        : t_mem_mosi;
  signal eth_serial_loopback : std_logic;

  -- ETH UDP data path from tx generation to rx verification
  signal udp_tx_sosi         : t_dp_sosi;
  signal udp_tx_siso         : t_dp_siso;

  signal udp_tx_pkt_sosi     : t_dp_sosi;
  signal udp_tx_pkt_siso     : t_dp_siso;

  signal udp_tx_hdr_pkt_sosi_arr : t_dp_sosi_arr(c_eth_nof_udp_ports - 1 downto 0);
  signal udp_tx_hdr_pkt_siso_arr : t_dp_siso_arr(c_eth_nof_udp_ports - 1 downto 0);

  signal udp_rx_frame_pkt_siso_arr : t_dp_siso_arr(c_eth_nof_udp_ports - 1 downto 0);
  signal udp_rx_frame_pkt_sosi_arr : t_dp_sosi_arr(c_eth_nof_udp_ports - 1 downto 0);

  signal udp_rx_pkt_siso     : t_dp_siso;
  signal udp_rx_pkt_sosi     : t_dp_sosi;

  signal udp_rx_siso         : t_dp_siso;
  signal udp_rx_sosi         : t_dp_sosi;

  -- ETH MM registers interface
  signal eth_reg_miso        : t_mem_miso;
  signal eth_reg_mosi        : t_mem_mosi;
  signal eth_reg_interrupt   : std_logic;

  signal eth_ram_miso        : t_mem_miso;
  signal eth_ram_mosi        : t_mem_mosi;

  signal dut_eth_init        : std_logic := '1';
  signal dut_tse_init        : std_logic := '1';
begin
  mm_clk <= (not mm_clk) or tb_end after c_mm_clk_period / 2;
  mm_rst <= '1', '0' after c_mm_clk_period * 20;

  st_clk <= (not st_clk) or tb_end after c_st_clk_period / 2;
  st_rst <= '1', '0' after c_st_clk_period * 7;

  eth_clk <= not eth_clk or tb_end after c_eth_clk_period / 2;  -- TSE reference clock

  random_0 <= func_common_random(random_0) when rising_edge(st_clk);
  random_1 <= func_common_random(random_1) when rising_edge(st_clk);

  proc_common_gen_duty_pulse(c_pulse_active, c_pulse_period,   '1', st_rst, st_clk, pulse_en, pulse_0);
  proc_common_gen_duty_pulse(c_pulse_active, c_pulse_period + 1, '1', st_rst, st_clk, pulse_en, pulse_1);

  ------------------------------------------------------------------------------
  -- STREAM CONTROL
  ------------------------------------------------------------------------------

  in_en          <= '1'                     when g_in_en = e_active      else
                    random_0(random_0'high) when g_in_en = e_random      else
                    pulse_0                 when g_in_en = e_pulse;

   udp_rx_siso.ready <= '1';
   udp_rx_siso.xon   <= '1';

  ------------------------------------------------------------------------------
  -- TSE SETUP
  ------------------------------------------------------------------------------
  p_tse_setup : process
  begin
    dut_tse_init <= '1';
    eth_tse_mosi.wr <= '0';
    eth_tse_mosi.rd <= '0';
     -- Wait for ETH init
    while dut_eth_init = '1' loop wait until rising_edge(mm_clk); end loop;
    -- Setup the TSE MAC
    proc_tech_tse_setup(c_technology_dut,
                        c_promis_en, c_tech_tse_tx_fifo_depth, c_tech_tse_rx_fifo_depth, c_tx_ready_latency,
                        c_dut_src_mac, eth_psc_access,
                        mm_clk, eth_tse_miso, eth_tse_mosi);
    dut_tse_init <= '0';
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- DATA GENERATION
  ------------------------------------------------------------------------------

  -- Generate data path input data
  p_stimuli : process
    variable v_sync      : std_logic := '0';
    variable v_bsn       : std_logic_vector(c_bsn_w - 1 downto 0) := (others => '0');
    variable v_symbol    : natural := c_symbol_init;
    variable v_channel   : natural := 1;
    variable v_err       : natural := c_err_init;

    variable v_mm_wr_addr : natural := 0;
    variable v_mm_wr_hdr  : std_logic_vector(c_word_w - 1 downto 0);
  begin
    udp_tx_sosi  <= c_dp_sosi_rst;
    reg_hdr_mosi <= c_mem_mosi_rst;
    ram_hdr_mosi <= c_mem_mosi_rst;
    eth_reg_mosi <= c_mem_mosi_rst;
    eth_ram_mosi <= c_mem_mosi_rst;

    dut_eth_init <= '1';

    proc_common_wait_until_low(mm_clk, mm_rst);
    proc_common_wait_some_cycles(mm_clk, 5);

    -- Set up the UDP demux
    proc_mem_mm_bus_wr(c_eth_reg_demux_wi + 0, c_dut_udp_port_en + c_dut_udp_port_st0, mm_clk, eth_reg_miso, eth_reg_mosi);  -- UDP port stream 0
    proc_common_wait_some_cycles(mm_clk, 5);

    -- Setup the RX config
    proc_mem_mm_bus_wr(c_eth_reg_config_wi + 0, c_dut_src_mac_lo,     mm_clk, eth_reg_miso, eth_reg_mosi);  -- control MAC address lo word
    proc_mem_mm_bus_wr(c_eth_reg_config_wi + 1, c_dut_src_mac_hi,     mm_clk, eth_reg_miso, eth_reg_mosi);  -- control MAC address hi halfword
    proc_mem_mm_bus_wr(c_eth_reg_config_wi + 2, c_dut_ip_addr,        mm_clk, eth_reg_miso, eth_reg_mosi);  -- control IP address
    proc_mem_mm_bus_wr(c_eth_reg_config_wi + 3, c_dut_udp_port_ctrl,  mm_clk, eth_reg_miso, eth_reg_mosi);  -- control UDP port
    -- Enable RX
    proc_mem_mm_bus_wr(c_eth_reg_control_wi + 0, c_dut_control_rx_en, mm_clk, eth_reg_miso, eth_reg_mosi);  -- control rx en
    dut_eth_init <= '0';

    -- MM Stimuli: write HEADER to RAM
    for i in c_network_total_header_32b_nof_words downto 1 loop
      -- Extract words from SLV from left to right
      v_mm_wr_hdr := c_total_hdr_slv(i * c_word_w - 1 downto i * c_word_w - c_word_w);
      proc_mem_mm_bus_wr(v_mm_wr_addr, v_mm_wr_hdr, mm_clk, ram_hdr_mosi);
      proc_common_wait_some_cycles(mm_clk, 5);

      if v_mm_wr_addr < c_network_total_header_32b_nof_words - 1 then
        v_mm_wr_addr := v_mm_wr_addr + 1;
      end if;
    end loop;

    -- Release the header onto the DP
    proc_mem_mm_bus_wr(0, 1, mm_clk, reg_hdr_mosi);

    -- Begin of ST stimuli
    proc_common_wait_until_low(st_clk, st_rst);
    proc_common_wait_some_cycles(st_clk, 50);

    for R in 0 to c_nof_repeat - 1 loop
      v_sync := sel_a_b(R mod c_sync_period = c_sync_offset, '1', '0');  -- v_bsn = R
      proc_dp_gen_block_data(c_rl, true, g_data_w, g_symbol_w, v_symbol, 0, 0, c_tb_nof_data, v_channel, v_err, v_sync, TO_DP_BSN(R), st_clk, in_en, udp_tx_siso, udp_tx_sosi);
      v_bsn     := INCR_UVEC(v_bsn, 1);
      v_symbol  := (v_symbol + c_tb_nof_data) mod c_symbol_mod;
      v_err     := v_err + 1;
      --proc_common_wait_some_cycles(st_clk, 10);               -- create gap between frames
    end loop;

    -- End of stimuli
    proc_common_wait_some_cycles(st_clk, 50);  -- depends on stream control
    verify_done <= '1';
    proc_common_wait_some_cycles(st_clk, 1);
    verify_done <= '0';

    -- Resync to MM clk
    proc_common_wait_some_cycles(mm_clk, 5);

    -- Read the stripped header via MM bus and print it in the transcript window
    print_str("Reading stripped header from RAM:");
    for i in 0 to c_network_total_header_32b_nof_words - 1 loop
      proc_mem_mm_bus_rd(i, mm_clk, ram_hdr_mosi);
      proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
      print_str("[" & time_to_str(now) & "] 0x" & slv_to_hex(ram_hdr_miso.rddata(c_word_w - 1 downto 0)));
    end loop;

    proc_common_wait_until_high(st_clk, out_gap);
    proc_common_wait_some_cycles(st_clk, 1000);
    tb_end <= '1';
    wait;
  end process;

  -- Stop the simulation
  p_tb_end : process
  begin
    wait until tb_end = '1';
    wait for 10 us;
    assert false report "Simulation tb_eth_udp_offload finished." severity NOTE;
    wait;
  end process;

  ------------------------------------------------------------------------------
  -- DATA VERIFICATION
  ------------------------------------------------------------------------------

  verify_en <= '1' when rising_edge(st_clk) and udp_rx_sosi.sop = '1';  -- verify enable after first output sop

  -- SOSI control
  proc_dp_verify_valid(c_rl, st_clk, verify_en, udp_rx_siso.ready, prev_udp_rx_ready, udp_rx_sosi.valid);  -- Verify that the output valid fits with the output ready latency
  proc_dp_verify_gap_invalid(st_clk, udp_rx_sosi.valid, udp_rx_sosi.sop, udp_rx_sosi.eop, out_gap);  -- Verify that the output valid is low between blocks from eop to sop

  -- SOSI data
  -- . verify that the output is incrementing symbols, like the input stimuli
  proc_dp_verify_symbols(c_rl, g_data_w, g_symbol_w, st_clk, verify_en, udp_rx_siso.ready, udp_rx_sosi.valid, udp_rx_sosi.eop, udp_rx_sosi.data(g_data_w - 1 downto 0), udp_rx_sosi.empty, prev_udp_rx_data);

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  dut : entity work.eth
  generic map (
    g_technology         => c_technology_dut,
    g_cross_clock_domain => true
  )
  port map (
    -- Clocks and reset
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    eth_clk           => eth_clk,
    st_rst            => st_rst,
    st_clk            => st_clk,
    -- UDP transmit interface
    -- . ST sink
    udp_tx_snk_in_arr  => udp_tx_hdr_pkt_sosi_arr,
    udp_tx_snk_out_arr => udp_tx_hdr_pkt_siso_arr,
    -- UDP receive interface
    -- . ST source
    udp_rx_src_in_arr  => udp_rx_frame_pkt_siso_arr,
    udp_rx_src_out_arr => udp_rx_frame_pkt_sosi_arr,
    -- Control Memory Mapped Slaves
    tse_sla_in        => eth_tse_mosi,
    tse_sla_out       => eth_tse_miso,
    reg_sla_in        => eth_reg_mosi,
    reg_sla_out       => eth_reg_miso,
    reg_sla_interrupt => eth_reg_interrupt,
    ram_sla_in        => eth_ram_mosi,
    ram_sla_out       => eth_ram_miso,
    -- PHY interface
    eth_txp           => eth_serial_loopback,
    eth_rxp           => eth_serial_loopback,
    -- LED interface
    tse_led           => open
  );

  u_hdr_insert : entity dp_lib.dp_hdr_insert
  generic map (
    g_data_w        => g_data_w,
    g_symbol_w      => g_symbol_w,
    g_hdr_nof_words => c_network_total_header_32b_nof_words
  )
  port map (
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,

    st_rst      => st_rst,
    st_clk      => st_clk,

    reg_mosi    => reg_hdr_mosi,
    ram_mosi    => ram_hdr_mosi,

    snk_out     => udp_tx_pkt_siso,
    snk_in      => udp_tx_pkt_sosi,

    src_in      => udp_tx_hdr_pkt_siso_arr(0),
    src_out     => udp_tx_hdr_pkt_sosi_arr(0)
  );

  u_dp_packet_enc : entity dp_lib.dp_packet_enc
  generic map (
    g_data_w => g_data_w
  )
  port map (
    rst       => st_rst,
    clk       => st_clk,

    snk_out   => udp_tx_siso,
    snk_in    => udp_tx_sosi,

    src_in    => udp_tx_pkt_siso,
    src_out   => udp_tx_pkt_sosi
  );

  u_frame_remove : entity dp_lib.dp_frame_remove
  generic map (
    g_data_w        => g_data_w,
    g_symbol_w      => g_symbol_w,
    g_hdr_nof_words => c_network_total_header_32b_nof_words,
    g_tail_nof_words => (c_network_eth_crc_len * g_symbol_w) / c_word_w
  )
  port map (
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,

    st_rst      => st_rst,
    st_clk      => st_clk,

    snk_out     => udp_rx_frame_pkt_siso_arr(0),
    snk_in      => udp_rx_frame_pkt_sosi_arr(0),

    sla_in      => ram_hdr_mosi,
    sla_out     => ram_hdr_miso,

    src_in      => udp_rx_pkt_siso,
    src_out     => udp_rx_pkt_sosi
  );

  u_dp_packet_dec : entity dp_lib.dp_packet_dec
  generic map (
    g_data_w => g_data_w
  )
  port map (
    rst       => st_rst,
    clk       => st_clk,

    snk_out   => udp_rx_pkt_siso,
    snk_in    => udp_rx_pkt_sosi,

    src_in    => udp_rx_siso,
    src_out   => udp_rx_sosi
  );
end tb;
