-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

library IEEE, common_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use work.i2c_smbus_pkg.all;
use work.i2c_dev_max1617_pkg.all;
use work.i2c_commander_pkg.all;

package i2c_dev_adu_pkg is
  -- I2C slave addresses of the devices on the I2C bus on ADU
  constant I2C_ADU_MAX1617_ADR     : natural := MAX1617_ADR_LOW_LOW;  -- temperature sensor, slave address is "0011000"
  constant I2C_ADU_PCA9555_ADR     : natural := 16#20#;  -- IO expander, slave address is "0100" & A[2:0], A[2:0]= "000"

  -- Commander protocol lists for ADU control

  -- * The S31* (for I2C_ADU_MAX1617_ADR) and S40* (for I2C_ADU_PCA9555_ADR)
  --   sequences were created with MATLAB by Gijs Schoonderbeek, they are
  --   commented here and column wise copy-pasted into the protocol list format
  --   using SMBUS_WRITE_WORD and SMBUS_WRITE_BYTE.

  -- * Temperature sensor:
  --       . c_i2c_adu_max1617_protocol_list_read_temp

  -- S310100P
  constant c_i2c_adu_max1617_expected_mask_read_temp   : std_logic_vector := c_i2c_cmdr_expected_mask_read_one_byte;
  constant c_i2c_adu_max1617_nof_result_data_read_temp : natural := c_i2c_cmdr_nof_result_data_read_one_byte;
  constant c_i2c_adu_max1617_protocol_list_read_temp   : t_nat_natural_arr := (
    SMBUS_READ_BYTE,  I2C_ADU_MAX1617_ADR, MAX1617_CMD_READ_REMOTE_TEMP,
    SMBUS_C_END
  );

  -- * IO expander:
  --     + IO set
  --       . c_i2c_adu_pca9555_protocol_list_set_atten_0_0dB
  --       . c_i2c_adu_pca9555_protocol_list_set_atten_1_0dB
  --       . c_i2c_adu_pca9555_protocol_list_set_atten_10_0dB
  --       . c_i2c_adu_pca9555_protocol_list_set_atten_10_6dB
  --       . c_i2c_adu_pca9555_protocol_list_set_cal
  --     + ADC bit-band control
  --       . c_i2c_adu_pca9555_protocol_list_set_adc
  --       . c_i2c_adu_pca9555_protocol_list_set_adc_tm

  -- S4002C040P
  -- S4002C140P
  -- S4002C040P
  constant c_i2c_adu_pca9555_expected_mask_set_atten_0_0dB : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_atten_0_0dB : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C1#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_C_END
  );

  -- S4002C040P
  -- S4002C440P
  -- S4002C040P
  constant c_i2c_adu_pca9555_expected_mask_set_atten_1_0dB : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_atten_1_0dB : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C4#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_C_END
  );

  -- S4002C040P
  -- S4002C540P
  -- S4002C040P
  constant c_i2c_adu_pca9555_expected_mask_set_atten_10_0dB : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_atten_10_0dB : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C5#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_C_END
  );

  -- S4002C04CP
  -- S4002C54CP
  -- S4002C04CP
  constant c_i2c_adu_pca9555_expected_mask_set_atten_10_6dB : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_atten_10_6dB : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#4C#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C5#, 16#4C#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#4C#,
    SMBUS_C_END
  );

  -- S4003C0P
  -- S400340P
  constant c_i2c_adu_pca9555_expected_mask_set_cal : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_cal : t_nat_natural_arr := (
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#03#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#03#, 16#40#,
    SMBUS_C_END
  );

  -- S40060000P
  -- S4002C040P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S4002C540P
  -- S4002C040P
  -- S4003C0P
  -- S400340P
  constant c_i2c_adu_pca9555_expected_mask_set_adc : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_adc : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#06#, 16#00#, 16#00#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C5#, 16#40#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#03#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#03#, 16#40#,
    SMBUS_C_END
  );

  -- S40060000P
  -- S4002C040P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400240P
  -- S400250P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S400260P
  -- S400270P
  -- S4002C0P
  -- S4002D0P
  constant c_i2c_adu_pca9555_expected_mask_set_adc_tm : std_logic_vector := c_i2c_cmdr_expected_mask_none;
  constant c_i2c_adu_pca9555_protocol_list_set_adc_tm : t_nat_natural_arr := (
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#06#, 16#00#, 16#00#,
    SMBUS_WRITE_WORD, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#40#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#50#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#60#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#70#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#C0#,
    SMBUS_WRITE_BYTE, I2C_ADU_PCA9555_ADR, 16#02#, 16#D0#,
    SMBUS_C_END
  );
end package;
