-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Testbench for ip_arria10_e3sge3_tse_sgmii_gx.
-- Description:
--   The tb is self checking based on that tx_pkt_cnt=rx_pkt_cnt must be true
--   at the tb_end.
-- Usage:
--   > as 10
--   > run -all

library IEEE, common_lib;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use common_lib.common_pkg.all;

entity tb_ip_arria10_e3sge3_tse_sgmii_gx is
end tb_ip_arria10_e3sge3_tse_sgmii_gx;

architecture tb of tb_ip_arria10_e3sge3_tse_sgmii_gx is
  constant sys_clk_period             : time := 10 ns;  -- 100 MHz
  constant eth_clk_period             : time :=  8 ns;  -- 125 MHz
  constant serial_clk_period          : time :=  800 ps;  -- 1250 MHz   ????
  constant cdr_clk_period             : time :=  8000 ps;  -- 125  MHz  ????

  constant c_tse_reg_addr_w           : natural := 8;  -- = max 256 MAC registers
  constant c_tse_byte_addr_w          : natural := c_tse_reg_addr_w + 2;
  constant c_tse_byte_addr_pcs_offset : natural := 16#200#;  -- table 4.8, 4.9 in ug_ethernet.pdf
  constant c_tse_data_w               : natural := c_word_w;  -- = 32

  constant c_tse_symbol_w             : natural := c_byte_w;  -- = 8
  constant c_tse_symbol_max           : natural := 2**c_tse_symbol_w - 1;  -- = 255
  constant c_tse_symbols_per_beat     : natural := c_tse_data_w / c_tse_symbol_w;  -- = 4

  constant c_tse_pcs_reg_addr_w       : natural := 5;  -- = max 32 PCS registers
  constant c_tse_pcs_halfword_addr_w  : natural := c_tse_pcs_reg_addr_w + 1;  -- table 4.17 in ug_ethernet.pdf
  constant c_tse_pcs_byte_addr_w      : natural := c_tse_pcs_reg_addr_w + 2;
  constant c_tse_pcs_data_w           : natural := c_halfword_w;  -- = 16;

  constant c_tse_empty_w              : natural := 2;
  constant c_tse_tx_error_w           : natural := 1;
  constant c_tse_rx_error_w           : natural := 6;
  constant c_tse_error_w              : natural := largest(c_tse_tx_error_w, c_tse_rx_error_w);
  constant c_tse_err_stat_w           : natural := 18;
  constant c_tse_frm_type_w           : natural := 4;

  constant c_tse_tx_fifo_depth        : natural := 256;  -- nof words for Tx FIFO
  constant c_tse_rx_fifo_depth        : natural := 256;  -- nof words for Rx FIFO

  constant c_tse_promis_en            : boolean := false;
  --CONSTANT c_tse_promis_en            : BOOLEAN := TRUE;

  constant c_tx_data_type             : natural := 1;  -- 0 = symbols, 1 = counter
  constant c_tx_ready_latency         : natural := 0;
  constant c_nof_tx_not_valid         : natural := 0;  -- when > 0 then pull tx valid low for c_nof_tx_not_valid beats during tx

  constant c_eth_dst_mac              : std_logic_vector(47 downto 0) := X"10FA01020300";
  constant c_eth_src_mac              : std_logic_vector(47 downto 0) := X"123456789ABC";  -- = 12-34-56-78-9A-BC
  constant c_eth_ethertype            : std_logic_vector(15 downto 0) := X"10FA";

  type t_mm_bus is record
    -- Master In Slave Out
    waitreq   : std_logic;
    rddata    : std_logic_vector(c_tse_data_w - 1 downto 0);
    -- Master Out Slave In
    address   : std_logic_vector(c_tse_byte_addr_w - 1 downto 0);
    wrdata    : std_logic_vector(c_tse_data_w - 1 downto 0);
    wr        : std_logic;
    rd        : std_logic;
  end record;

  procedure proc_dbg_mm_bus(signal mm_miso  : in  t_mm_bus;
                            signal mm_mosi  : in  t_mm_bus;
                            signal dbg_mm   : out t_mm_bus) is
  begin
    dbg_mm.waitreq <= mm_miso.waitreq;
    dbg_mm.rddata  <= mm_miso.rddata;
    dbg_mm.address <= mm_mosi.address;
    dbg_mm.wrdata  <= mm_mosi.wrdata;
    dbg_mm.wr      <= mm_mosi.wr;
    dbg_mm.rd      <= mm_mosi.rd;
  end proc_dbg_mm_bus;

  -- Wait for MM access (either read or write) finished
  procedure proc_mm_access(signal mm_clk      : in  std_logic;
                           signal mm_waitreq  : in  std_logic;
                           signal mm_access   : out std_logic) is
  begin
    mm_access <= '1';
    wait until rising_edge(mm_clk);
    while mm_waitreq = '1' loop
      wait until rising_edge(mm_clk);
    end loop;
    mm_access <= '0';
  end proc_mm_access;

  -- Use word addressing for MAC registers according to table 4.8, 4.9
  procedure proc_wr_mac(constant mac_addr : in  natural;
                        constant mac_data : in  natural;
                        signal   mm_clk   : in  std_logic;
                        signal   mm_miso  : in  t_mm_bus;
                        signal   mm_mosi  : out t_mm_bus) is
  begin
    mm_mosi.address <= std_logic_vector(to_unsigned(mac_addr, c_tse_byte_addr_w));
    mm_mosi.wrdata  <= std_logic_vector(to_unsigned(mac_data, c_tse_data_w));
    proc_mm_access(mm_clk, mm_miso.waitreq, mm_mosi.wr);
  end proc_wr_mac;

  procedure proc_rd_mac(constant mac_addr : in  natural;
                        signal   mac_data : out natural;
                        signal   mm_clk   : in  std_logic;
                        signal   mm_miso  : in  t_mm_bus;
                        signal   mm_mosi  : out t_mm_bus) is
  begin
    mm_mosi.address <= std_logic_vector(to_unsigned(mac_addr, c_tse_byte_addr_w));
    proc_mm_access(mm_clk, mm_miso.waitreq, mm_mosi.rd);
    MAC_data <= to_integer(unsigned(mm_miso.rddata));
  end proc_rd_mac;

  -- Use halfword addressing for PCS register to match table 4.17
  procedure proc_wr_pcs(constant pcs_addr : in  natural;
                        constant pcs_data : in  natural;
                        signal   mm_clk   : in  std_logic;
                        signal   mm_miso  : in  t_mm_bus;
                        signal   mm_mosi  : out t_mm_bus) is
  begin
    mm_mosi.address                             <= std_logic_vector(to_unsigned(pcs_addr * 2 + c_tse_byte_addr_pcs_offset, c_tse_byte_addr_w));
    mm_mosi.wrdata                              <= (others => '0');
    mm_mosi.wrdata(c_tse_pcs_data_w - 1 downto 0) <= std_logic_vector(to_unsigned(pcs_data, c_tse_pcs_data_w));
    proc_mm_access(mm_clk, mm_miso.waitreq, mm_mosi.wr);
  end proc_wr_pcs;

  procedure proc_rd_pcs(constant pcs_addr : in  natural;
                        signal   pcs_data : out natural;
                        signal   mm_clk   : in  std_logic;
                        signal   mm_miso  : in  t_mm_bus;
                        signal   mm_mosi  : out t_mm_bus) is
  begin
    mm_mosi.address <= std_logic_vector(to_unsigned(pcs_addr * 2 + c_tse_byte_addr_pcs_offset, c_tse_byte_addr_w));
    proc_mm_access(mm_clk, mm_miso.waitreq, mm_mosi.rd);
    pcs_data <= to_integer(unsigned(mm_miso.rddata(c_tse_pcs_data_w - 1 downto 0)));
  end proc_rd_pcs;

  type t_tse_stream is record
    -- Source In or Sink Out
    ready    : std_logic;
    -- Source Out or Sink In
    data     : std_logic_vector(c_tse_data_w - 1 downto 0);
    valid    : std_logic;
    sop      : std_logic;
    eop      : std_logic;
    empty    : std_logic_vector(c_tse_empty_w - 1 downto 0);
    err      : std_logic_vector(c_tse_error_w - 1 downto 0);
  end record;

  procedure proc_dbg_tse_stream_src(signal src_in  : in  t_tse_stream;
                                    signal src_out : in  t_tse_stream;
                                    signal dbg_src : out t_tse_stream) is
  begin
    dbg_src.ready <= src_in.ready;
    dbg_src.data  <= src_out.data;
    dbg_src.valid <= src_out.valid;
    dbg_src.sop   <= src_out.sop;
    dbg_src.eop   <= src_out.eop;
    dbg_src.empty <= src_out.empty;
    dbg_src.err   <= src_out.err;
  end proc_dbg_tse_stream_src;

  procedure proc_dbg_tse_stream_snk(signal snk_in  : in  t_tse_stream;
                                    signal snk_out : in  t_tse_stream;
                                    signal dbg_snk : out t_tse_stream) is
  begin
    dbg_snk.ready <= snk_out.ready;
    dbg_snk.data  <= snk_in.data;
    dbg_snk.valid <= snk_in.valid;
    dbg_snk.sop   <= snk_in.sop;
    dbg_snk.eop   <= snk_in.eop;
    dbg_snk.empty <= snk_in.empty;
    dbg_snk.err   <= snk_in.err;
  end proc_dbg_tse_stream_snk;

  -- Handle TX ready
  -- Only support tx_ready_latency=0 or 1, corresponding to TX_ALMOST_FULL=3 or 4
  -- Support for tx_ready_latency>1 requires keeping previous ready information
  -- in a STD_LOGIC_VECTOR(tx_ready_latency-1 DOWNTO 0).
  procedure proc_ready_latency(constant c_latency : in  natural;
                               signal   clk       : in  std_logic;
                               signal   ready     : in  std_logic;
                               constant c_valid   : in  std_logic;
                               constant c_sop     : in  std_logic;
                               constant c_eop     : in  std_logic;
                               signal   out_valid : out std_logic;
                               signal   out_sop   : out std_logic;
                               signal   out_eop   : out std_logic) is
  begin
    if c_latency = 0 then
      out_valid <= c_valid;
      out_sop   <= c_sop;
      out_eop   <= c_eop;
      wait until rising_edge(clk);
      while ready /= '1' loop
        wait until rising_edge(clk);
      end loop;
    end if;
    if c_latency = 1 then
      while ready /= '1' loop
        out_valid <= '0';
        out_sop   <= '0';
        out_eop   <= '0';
        wait until rising_edge(clk);
      end loop;
      out_valid <= c_valid;
      out_sop   <= c_sop;
      out_eop   <= c_eop;
      wait until rising_edge(clk);
    end if;
  end proc_ready_latency;

  -- Transmit user packet
  -- . Use word aligned payload data, so with half word inserted before the 14 byte header
  -- . Packets can be send immediately after eachother so new sop directly after last eop
  -- . The word rate is controlled by respecting ready from the MAC
  procedure proc_tx_packet(constant dst_mac_addr : in  std_logic_vector(c_eth_dst_mac'range);
                           constant src_mac_addr : in  std_logic_vector(c_eth_src_mac'range);
                           constant ethertype    : in  std_logic_vector(c_eth_ethertype'range);
                           constant data_len     : in  natural;  -- in symbols = octets = bytes
                           signal   dp_clk       : in  std_logic;
                           signal   dp_src_in    : in  t_tse_stream;
                           signal   dp_src_out   : out t_tse_stream) is
    constant c_mod            : natural := data_len mod c_tse_symbols_per_beat;
    constant c_nof_data_beats : natural := data_len   / c_tse_symbols_per_beat + sel_a_b(c_mod, 1, 0);
    constant c_empty          : natural := sel_a_b(c_mod, c_tse_symbols_per_beat - c_mod, 0);
    variable v_sym            : unsigned(c_tse_symbol_w - 1 downto 0) := (others => '0');
    variable v_num            : unsigned(c_tse_data_w - 1 downto 0) := (others => '0');
  begin
    -- DST MAC
    dp_src_out.empty <= std_logic_vector(to_unsigned(0, c_tse_empty_w));
    dp_src_out.data              <= (others => '0');
    dp_src_out.data(15 downto 0) <= hton(dst_mac_addr(15 downto 0));  -- send to itself
    proc_ready_latency(c_tx_ready_latency, dp_clk, dp_src_in.ready, '1', '1', '0', dp_src_out.valid, dp_src_out.sop, dp_src_out.eop);
    dp_src_out.data  <= hton(dst_mac_addr(47 downto 16));
    proc_ready_latency(c_tx_ready_latency, dp_clk, dp_src_in.ready, '1', '0', '0', dp_src_out.valid, dp_src_out.sop, dp_src_out.eop);
    -- SRC MAC
    dp_src_out.data  <= hton(src_mac_addr(31 downto 0));
    proc_ready_latency(c_tx_ready_latency, dp_clk, dp_src_in.ready, '1', '0', '0', dp_src_out.valid, dp_src_out.sop, dp_src_out.eop);
    -- SRC MAC & ETHERTYPE
    dp_src_out.data  <= hton(src_mac_addr(47 downto 32)) & hton(c_eth_ethertype);
    -- DATA
    for I in 0 to c_nof_data_beats - 1 loop
      proc_ready_latency(c_tx_ready_latency, dp_clk, dp_src_in.ready, '1', '0', '0', dp_src_out.valid, dp_src_out.sop, dp_src_out.eop);
      if c_tx_data_type = 0 then
        -- data : X"01020304", X"05060708", X"090A0B0C", etc
        for J in c_tse_symbols_per_beat - 1 downto 0 loop
          v_sym := v_sym + 1;
          dp_src_out.data((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w) <= std_logic_vector(v_sym);
        end loop;
      else
        -- data : X"00000001", X"00000002", X"00000003", etc
        v_num := v_num + 1;
        dp_src_out.data <= std_logic_vector(v_num);
      end if;
      -- tb : pull valid low for some time during the middle of the payload
      if c_nof_tx_not_valid > 0 and I = c_nof_data_beats / 2 then
        dp_src_out.valid <= '0';
        for I in 0 to c_nof_tx_not_valid loop wait until rising_edge(dp_clk); end loop;
        dp_src_out.valid <= '1';
      end if;
    end loop;
    if c_empty > 0 then
      dp_src_out.empty <= std_logic_vector(to_unsigned(c_empty, c_tse_empty_w));
      for J in c_empty - 1 downto 0 loop
        dp_src_out.data((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w) <= (others => '0');
      end loop;
    end if;
    proc_ready_latency(c_tx_ready_latency, dp_clk, dp_src_in.ready, '1', '0', '1', dp_src_out.valid, dp_src_out.sop, dp_src_out.eop);
    dp_src_out.data  <= (others => '0');
    dp_src_out.valid <= '0';
    dp_src_out.eop   <= '0';
    dp_src_out.empty <= std_logic_vector(to_unsigned(0, c_tse_empty_w));
  end proc_tx_packet;

  procedure proc_valid_sop(signal   clk       : in  std_logic;
                           signal   in_valid  : in  std_logic;
                           signal   in_sop    : in  std_logic) is
  begin
    wait until rising_edge(clk);
    while in_valid /= '1' and in_sop /= '1' loop
      wait until rising_edge(clk);
    end loop;
  end proc_valid_sop;

  procedure proc_valid(signal   clk       : in  std_logic;
                       signal   in_valid  : in  std_logic) is
  begin
    wait until rising_edge(clk);
    while in_valid /= '1' loop
      wait until rising_edge(clk);
    end loop;
  end proc_valid;

  -- Receive packet
  -- . Use word aligned payload data, so with half word inserted before the 14 byte header
  -- . Packets can be always be received, assume the user application is always ready
  -- . The CRC32 is also passed on to the user at eop.
  -- . Note that when empty/=0 then the CRC32 is not word aligned, so therefore use prev_data to be able
  --   to handle part of last data word in case empty/=0 at eop
  procedure proc_rx_packet(constant dst_mac_addr : in  std_logic_vector(c_eth_dst_mac'range);
                           constant src_mac_addr : in  std_logic_vector(c_eth_src_mac'range);
                           constant ethertype    : in  std_logic_vector(c_eth_ethertype'range);
                           signal   dp_clk       : in  std_logic;
                           signal   dp_snk_in    : in  t_tse_stream;
                           signal   dp_snk_out   : out t_tse_stream) is
    variable v_sym            : unsigned(c_tse_symbol_w - 1 downto 0) := (others => '0');
    variable v_num            : unsigned(c_tse_data_w - 1 downto 0) := (others => '0');
    variable v_empty          : natural;
    variable v_first          : boolean := true;
    variable v_data           : std_logic_vector(c_tse_data_w - 1 downto 0);
    variable v_prev_data      : std_logic_vector(c_tse_data_w - 1 downto 0);
  begin
    -- Keep ff_rx_snk_out.ready='1' all the time
    dp_snk_out.ready <= '1';
    -- Verify DST MAC
    proc_valid_sop(dp_clk, dp_snk_in.valid, dp_snk_in.sop);
    assert dp_snk_in.data(31 downto 16) = X"0000"                           report "RX: Alignment half word not zero" severity ERROR;
    assert dp_snk_in.data(15 downto  0) = hton(dst_mac_addr(15 downto 0))   report "RX: Wrong dst_mac_addr(15 downto 0)" severity ERROR;
    proc_valid(dp_clk, dp_snk_in.valid);
    assert dp_snk_in.data(31 downto  0) = hton(dst_mac_addr(47 downto 16))  report "RX: Wrong dst_mac_addr(47 downto 16)" severity ERROR;
    -- Verify SRC MAC
    proc_valid(dp_clk, dp_snk_in.valid);
    assert dp_snk_in.data(31 downto  0) = hton(src_mac_addr(31 downto 0))   report "RX: Wrong src_mac_addr(31 downto 0)" severity ERROR;
    -- Verify SRC MAC & ETHERTYPE
    proc_valid(dp_clk, dp_snk_in.valid);
    assert dp_snk_in.data(31 downto 16) = hton(src_mac_addr(47 downto 32))  report "RX: Wrong src_mac_addr(47 downto 32)" severity ERROR;
    assert dp_snk_in.data(15 downto  0) = hton(c_eth_ethertype)             report "RX: Wrong ethertype" severity ERROR;
    -- Verify DATA
    v_first := true;
    proc_valid(dp_clk, dp_snk_in.valid);
    while dp_snk_in.eop /= '1' loop
      v_prev_data := v_data;
      v_data      := dp_snk_in.data;
      if v_first = false then
        if c_tx_data_type = 0 then
          -- data : X"01020304", X"05060708", X"090A0B0C", etc
          for J in c_tse_symbols_per_beat - 1 downto 0 loop
            v_sym := v_sym + 1;
            assert unsigned(v_prev_data((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w)) = v_sym report "RX: Wrong data symbol" severity ERROR;
          end loop;
        else
          -- data : X"00000001", X"00000002", X"00000003", etc
          v_num := v_num + 1;
          assert unsigned(v_prev_data) = v_num report "RX: Wrong data word" severity ERROR;
        end if;
      end if;
      v_first := false;
      proc_valid(dp_clk, dp_snk_in.valid);
    end loop;
    -- Verify last DATA and CRC32 if empty /=0 else the last word is only the CRC32
    v_prev_data := v_data;
    v_data      := dp_snk_in.data;
    v_empty     := to_integer(unsigned(dp_snk_in.empty));
    if v_empty > 0 then
      for J in v_empty - 1 downto 0 loop
        v_prev_data((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w) := (others => '0');
      end loop;
      if c_tx_data_type = 0 then
        -- data : X"01020304", X"05060708", X"090A0B0C", etc
        for J in c_tse_symbols_per_beat - 1 downto v_empty loop  -- ignore CRC32 symbols in last data word
          v_sym := v_sym + 1;
          assert unsigned(v_prev_data((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w)) = v_sym report "RX: Wrong empty data symbol" severity ERROR;
        end loop;
      else
        -- data : X"00000001", X"00000002", X"00000003", etc
        v_num := v_num + 1;
        for J in v_empty - 1 downto 0 loop
          v_num((J + 1) * c_tse_symbol_w - 1 downto J * c_tse_symbol_w) := (others => '0');  -- force CRC32 symbols in last data word to 0
        end loop;
        assert unsigned(v_prev_data) = v_num report "RX: Wrong empty data word" severity ERROR;
      end if;
    else
      -- No verify on CRC32 word
    end if;
  end proc_rx_packet;

  -- Clocks and reset
  signal tb_end         : std_logic := '0';
  signal eth_clk        : std_logic := '0';
  signal sys_clk        : std_logic := '0';
  signal dp_clk         : std_logic;
  signal mm_clk         : std_logic;
  signal mm_rst         : std_logic;

  -- TSE MAC control interface
  signal mm_init        : std_logic := '1';
  signal mm_miso        : t_mm_bus;  -- master in  slave out
  signal mm_mosi        : t_mm_bus;  -- master out slave in

  signal pcs_rddata     : natural;  -- [c_tse_pcs_data_w-1:0]

  signal tse_led_an     : std_logic;
  signal tse_led_link   : std_logic;

  -- TSE MAC transmit interface
  -- . Avalon ST source
  signal ff_tx_src_in   : t_tse_stream;
  signal ff_tx_src_out  : t_tse_stream;
  -- . MAC specific
  signal ff_tx_crc_fwd  : std_logic;
  signal ff_tx_septy    : std_logic;
  signal ff_tx_a_full   : std_logic;
  signal ff_tx_a_empty  : std_logic;
  signal ff_tx_uflow    : std_logic;

  -- TSE MAC receive interface
  -- . Avalon ST sink
  signal ff_rx_snk_in   : t_tse_stream;
  signal ff_rx_snk_out  : t_tse_stream;
  -- . MAC specific
  signal ff_rx_ethertype: std_logic_vector(c_tse_err_stat_w - 1 downto 0);
  signal ff_rx_frm_type : std_logic_vector(c_tse_frm_type_w - 1 downto 0);
  signal ff_rx_dsav     : std_logic;
  signal ff_rx_a_full   : std_logic;
  signal ff_rx_a_empty  : std_logic;

  -- TSE PHY interface
  signal eth_txp        : std_logic;
  signal eth_rxp        : std_logic;

  -- TSE PHY GX
  signal tx_serial_clk  : std_logic_vector(0 downto 0) := (others => '0');
  signal rx_cdr_refclk  : std_logic := '0';

  -- Verification
  signal tx_pkt_cnt     : natural := 0;
  signal rx_pkt_cnt     : natural := 0;

  -- Debug signals to combine valid in and out of records
  signal dbg_mm         : t_mm_bus;
  signal dbg_ff_tx      : t_tse_stream;
  signal dbg_ff_rx      : t_tse_stream;
begin
  eth_clk <= not eth_clk after eth_clk_period / 2;  -- TSE reference clock
  sys_clk <= not sys_clk after sys_clk_period / 2;  -- System clock

  mm_clk  <= sys_clk;
  dp_clk  <= sys_clk;

  -- Debug signals to combine valid in and out of records
  proc_dbg_mm_bus(        mm_miso,      mm_mosi,       dbg_mm);
  proc_dbg_tse_stream_src(ff_tx_src_in, ff_tx_src_out, dbg_ff_tx);
  proc_dbg_tse_stream_snk(ff_rx_snk_in, ff_rx_snk_out, dbg_ff_rx);

  -- run 1 us
  p_mm_stimuli : process
  begin
    mm_init  <= '1';
    mm_mosi.wr <= '0';
    mm_mosi.rd <= '0';

    -- reset release
    mm_rst <= '1';
    for I in 0 to 9 loop wait until rising_edge(mm_clk); end loop;
    mm_rst <= '0';
    for I in 0 to 9 loop wait until rising_edge(mm_clk); end loop;

    -- PSC control
    proc_rd_pcs(16#22#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- REV --> 0x0901
    proc_wr_pcs(16#28#,   16#0008#, mm_clk, mm_miso, mm_mosi);  -- IF_MODE <-- Force 1GbE, no autonegatiation
    proc_rd_pcs(16#00#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- CONTROL --> 0x1140
    proc_rd_pcs(16#02#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- STATUS --> 0x000D
    proc_wr_pcs(16#00#,   16#0140#, mm_clk, mm_miso, mm_mosi);  -- CONTROL <-- Auto negotiate disable

    -- MAC control
    proc_rd_mac(16#000#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- REV --> CUST_VERSION & 0x0901
    if c_tse_promis_en = false then
      proc_wr_mac(16#008#, 16#0100004B#, mm_clk, mm_miso, mm_mosi);
    else
      proc_wr_mac(16#008#, 16#0100005B#, mm_clk, mm_miso, mm_mosi);
    end if;
      -- COMMAND_CONFIG <--
      -- Only the bits relevant to UniBoard are explained here, others are 0
      -- [    0] = TX_ENA             = 1, enable tx datapath
      -- [    1] = RX_ENA             = 1, enable rx datapath
      -- [    2] = XON_GEN            = 0
      -- [    3] = ETH_SPEED          = 1, enable 1GbE operation
      -- [    4] = PROMIS_EN          = 0, when 1 then receive all frames
      -- [    5] = PAD_EN             = 0, when 1 enable receive padding removal (requires ethertype=payload length)
      -- [    6] = CRC_FWD            = 1, enable receive CRC forward
      -- [    7] = PAUSE_FWD          = 0
      -- [    8] = PAUSE_IGNORE       = 0
      -- [    9] = TX_ADDR_INS        = 0, when 1 then MAX overwrites tx SRC MAC with mac_0,1 or one of the supplemental mac
      -- [   10] = HD_ENA             = 0
      -- [   11] = EXCESS_COL         = 0
      -- [   12] = LATE_COL           = 0
      -- [   13] = SW_RESET           = 0, when 1 MAC disables tx and rx, clear statistics and flushes receive FIFO
      -- [   14] = MHAS_SEL           = 0, select multicast address resolutions hash-code mode
      -- [   15] = LOOP_ENA           = 0
      -- [18-16] = TX_ADDR_SEL[2:0]   = 000, TX_ADDR_INS insert mac_0,1 or one of the supplemental mac
      -- [   19] = MAGIC_EN           = 0
      -- [   20] = SLEEP              = 0
      -- [   21] = WAKEUP             = 0
      -- [   22] = XOFF_GEN           = 0
      -- [   23] = CNT_FRM_ENA        = 0
      -- [   24] = NO_LGTH_CHECK      = 1, when 0 then check payload length of received frames (requires ethertype=payload length)
      -- [   25] = ENA_10             = 0
      -- [   26] = RX_ERR_DISC        = 0, when 1 then discard erroneous frames (requires store and forward mode, so rx_section_full=0)
      --                                   when 0 then pass on with rx_err[0]=1
      -- [   27] = DISABLE_RD_TIMEOUT = 0
      -- [30-28] = RSVD               = 000
      -- [   31] = CNT_RESET          = 0, when 1 clear statistics
    proc_wr_mac(16#00C#, 16#56789ABC#, mm_clk, mm_miso, mm_mosi);  -- MAC_0
    proc_wr_mac(16#010#, 16#00001234#, mm_clk, mm_miso, mm_mosi);  -- MAC_1 <-- SRC_MAC = 12-34-56-78-9A-BC
    proc_wr_mac(16#05C#, 16#0000000C#, mm_clk, mm_miso, mm_mosi);  -- TX_IPG_LENGTH <-- interpacket gap = 12
    proc_wr_mac(16#014#, 16#000005EE#, mm_clk, mm_miso, mm_mosi);  -- FRM_LENGTH <-- receive max frame length = 1518

    -- FIFO legenda:
    -- . Tx section full  = There is enough data in the FIFO to start reading it, when 0 then store and forward.
    -- . Rx section full  = There is enough data in the FIFO to start reading it, when 0 then store and forward.
    -- . Tx section empty = There is not much empty space anymore in the FIFO, warn user via ff_tx_septy
    -- . Rx section empty = There is not much empty space anymore in the FIFO, inform remote device via XOFF flow control
    -- . Tx almost full   = Assert ff_tx_a_full and deassert ff_tx_rdy. Furthermore TX_ALMOST_FULL = c_tx_ready_latency+3,
    --                      so choose 3 for zero tx ready latency
    -- . Rx almost full   = Assert ff_rx_a_full and if the user is not ready ff_rx_rdy then:
    --                      --> break off the reception with an error to avoid FIFO overflow
    -- . Tx almost empty  = Assert ff_tx_a_empty and if the FIFO does not contain a eop yet then:
    --                      --> break off the transmission with an error to avoid FIFO underflow
    -- . Rx almost empty  = Assert ff_rx_a_empty
    -- Typical FIFO values:
    -- . TX_SECTION_FULL  = 16   > 8   = TX_ALMOST_EMPTY
    -- . RX_SECTION_FULL  = 16   > 8   = RX_ALMOST_EMPTY
    -- . TX_SECTION_EMPTY = D-16 < D-3 = Tx FIFO depth - TX_ALMOST_FULL
    -- . RX_SECTION_EMPTY = D-16 < D-8 = Rx FIFO depth - RX_ALMOST_FULL
    -- . c_tse_tx_fifo_depth = 1 M9K = 256*32b = 1k * 8b is sufficient when the Tx user respects ff_tx_rdy, to store a complete
    --                         ETH packet would require 1518 byte, so 2 M9K = 2k * 8b
    -- . c_tse_rx_fifo_depth = 1 M9K = 256*32b = 1k * 8b is sufficient when the Rx user ff_rx_rdy is sufficiently active
    proc_wr_mac(16#01C#, c_tse_rx_fifo_depth - 16, mm_clk, mm_miso, mm_mosi);  -- RX_SECTION_EMPTY <-- default FIFO depth - 16, >3
    proc_wr_mac(16#020#,                     16, mm_clk, mm_miso, mm_mosi);  -- RX_SECTION_FULL  <-- default 16
    proc_wr_mac(16#024#, c_tse_tx_fifo_depth - 16, mm_clk, mm_miso, mm_mosi);  -- TX_SECTION_EMPTY <-- default FIFO depth - 16, >3
    proc_wr_mac(16#028#,                     16, mm_clk, mm_miso, mm_mosi);  -- TX_SECTION_FULL  <-- default 16, >~ 8 otherwise no tx
    proc_wr_mac(16#02C#,                      8, mm_clk, mm_miso, mm_mosi);  -- RX_ALMOST_EMPTY  <-- default 8
    proc_wr_mac(16#030#,                      8, mm_clk, mm_miso, mm_mosi);  -- RX_ALMOST_FULL   <-- default 8
    proc_wr_mac(16#034#,                      8, mm_clk, mm_miso, mm_mosi);  -- TX_ALMOST_EMPTY  <-- default 8
    proc_wr_mac(16#038#,   c_tx_ready_latency + 3, mm_clk, mm_miso, mm_mosi);  -- TX_ALMOST_FULL   <-- default 3

    proc_rd_mac(16#0E8#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- TX_CMD_STAT --> 0x00040000 : [18]=1 TX_SHIFT16, [17]=0 OMIT_CRC
    proc_rd_mac(16#0EC#, pcs_rddata, mm_clk, mm_miso, mm_mosi);  -- RX_CMD_STAT --> 0x02000000 : [25]=1 RX_SHIFT16

    wait until rising_edge(mm_clk);
    mm_init <= '0';

    wait;
  end process;

  p_tx_frame : process
  begin
    -- . Avalon ST
    ff_tx_src_out.data  <= (others => '0');
    ff_tx_src_out.valid <= '0';
    ff_tx_src_out.sop   <= '0';
    ff_tx_src_out.eop   <= '0';
    ff_tx_src_out.empty <= (others => '0');
    ff_tx_src_out.err   <= (others => '0');
    -- . MAC specific
    ff_tx_crc_fwd <= '0';

    while mm_init /= '0' loop
      wait until rising_edge(dp_clk);
    end loop;
    while tse_led_link /= '1' loop
      wait until rising_edge(dp_clk);
    end loop;
    for I in 0 to 9 loop wait until rising_edge(dp_clk); end loop;

--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac, "0000000000010000",   16, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype,   16, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 16, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
    proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1499, dp_clk, ff_tx_src_in, ff_tx_src_out);  -- verify st empty
    proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);
--     proc_tx_packet(c_eth_src_mac, c_eth_src_mac,    c_eth_ethertype, 1500, dp_clk, ff_tx_src_in, ff_tx_src_out);

    for I in 0 to 1500 * 4 loop wait until rising_edge(dp_clk); end loop;
    tb_end <= '1';
    wait;
  end process;

  p_rx_frame : process
  begin
    -- . Avalon ST
    ff_rx_snk_out.ready <= '0';

    while mm_init /= '0' loop
      wait until rising_edge(dp_clk);
    end loop;

    -- Receive forever
    while true loop
      proc_rx_packet(c_eth_src_mac, c_eth_src_mac, c_eth_ethertype, dp_clk, ff_rx_snk_in, ff_rx_snk_out);
    end loop;

    wait;
  end process;

  dut : entity work.ip_arria10_e3sge3_tse_sgmii_gx
    -- The ip_arria10_e3sge3_tse_sgmii_gx needs to be regenerated if its parameters are changed.
    -- . ENABLE_SHIFT16  = 1   : Align packet headers to 32 bit, useful for Nios data handling
    -- . ENABLE_SUP_ADDR = 0   : An extra MAC addresses can e.g. be used as service MAC for tests
    -- . ENA_HASH        = 0   : A multi cast hash table can be used to address all nodes at once
    -- . STAT_CNT_ENA    = 0   : PHY statistics counts are useful for monitoring, but not realy needed
    -- . EG_FIFO         = 256 : Tx FIFO depth in nof 32 bit words (256 --> 1 M9K)
    -- . ING_FIFO        = 256 : Rx FIFO depth in nof 32 bit words (256 --> 1 M9K)
  port map (
    -- MAC transmit interface
    -- . Avalon ST
    ff_tx_clk      => dp_clk,  -- : in  std_logic                     := '0';             --     transmit_clock_connection.clk
    ff_tx_rdy      => ff_tx_src_in.ready,  -- : out std_logic;                                        --                              .ready
    ff_tx_data     => ff_tx_src_out.data,  -- : in  std_logic_vector(31 downto 0) := (others => '0'); --                      transmit.data
    ff_tx_wren     => ff_tx_src_out.valid,  -- : in  std_logic                     := '0';             --                              .valid
    ff_tx_sop      => ff_tx_src_out.sop,  -- : in  std_logic                     := '0';             --                              .startofpacket
    ff_tx_eop      => ff_tx_src_out.eop,  -- : in  std_logic                     := '0';             --                              .endofpacket
    ff_tx_mod      => ff_tx_src_out.empty,  -- : in  std_logic_vector(1 downto 0)  := (others => '0'); --                              .empty
    ff_tx_err      => ff_tx_src_out.err(0),  -- : in  std_logic                     := '0';             --                              .error
    -- . MAC specific
    ff_tx_crc_fwd  => ff_tx_crc_fwd,  -- : in  std_logic                     := '0';             --           mac_misc_connection.ff_tx_crc_fwd     -- when '0' MAC inserts CRC32 after eop
    ff_tx_septy    => ff_tx_septy,  -- : out std_logic;                                        --                              .ff_tx_septy       -- when '0' then tx FIFO goes above section-empty threshold
    ff_tx_a_full   => ff_tx_a_full,  -- : out std_logic;                                        --                              .ff_tx_a_full      -- when '1' then tx FIFO goes above almost-full threshold
    ff_tx_a_empty  => ff_tx_a_empty,  -- : out std_logic;                                        --                              .ff_tx_a_empty     -- when '1' then tx FIFO goes below almost-empty threshold
    tx_ff_uflow    => ff_tx_uflow,  -- : out std_logic;                                        --                              .tx_ff_uflow       -- when '1' then tx FIFO underflow
    -- MAC receive interface
    -- . Avalon STs
    ff_rx_clk      => dp_clk,  -- : in  std_logic                     := '0';             --      receive_clock_connection.clk
    ff_rx_rdy      => ff_rx_snk_out.ready,  -- : in  std_logic                     := '0';             --                              .ready
    ff_rx_data     => ff_rx_snk_in.data,  -- : out std_logic_vector(31 downto 0);                    --                       receive.data
    ff_rx_dval     => ff_rx_snk_in.valid,  -- : out std_logic;                                        --                              .valid
    ff_rx_sop      => ff_rx_snk_in.sop,  -- : out std_logic;                                        --                              .startofpacket
    ff_rx_eop      => ff_rx_snk_in.eop,  -- : out std_logic;                                        --                              .endofpacket
    ff_rx_mod      => ff_rx_snk_in.empty,  -- : out std_logic_vector(1 downto 0);                     --                              .empty
    rx_err         => ff_rx_snk_in.err,  -- : out std_logic_vector(5 downto 0);                     --                              .error
                                            -- [5] collision error (can only occur in half duplex mode)
                                            -- [4] PHY error on GMII
                                            -- [3] receive frame truncated due to FIFO overflow
                                            -- [2] CRC-32 error
                                            -- [1] invalid length
                                            -- [0] = OR of [1:5]
    -- . MAC specific
    rx_err_stat    => ff_rx_ethertype,  -- : out std_logic_vector(17 downto 0);                    --                              .rx_err_stat      -- [17,16] VLAN info, [15:0] Ethernet lentgh/type field
    rx_frm_type    => ff_rx_frm_type,  -- : out std_logic_vector(3 downto 0);                     --                              .rx_frm_type      -- [3]=VLAN, [2]=Broadcast, [1]=Multicast, [0]=Unicast
    ff_rx_dsav     => ff_rx_dsav,  -- : out std_logic;                                        --                              .ff_rx_dsav       -- rx frame available, but not necessarily a complete frame
    ff_rx_a_full   => ff_rx_a_full,  -- : out std_logic;                                        --                              .ff_rx_a_full     -- when '1' then rx FIFO goes above almost-full threshold
    ff_rx_a_empty  => ff_rx_a_empty,  -- : out std_logic;                                        --                              .ff_rx_a_empty    -- when '1' sthen rx FIFO goes below almost-empty threshold
    -- Reset
    reset          => mm_rst,  -- : in  std_logic                     := '0';             --              reset_connection.reset            -- asynchronous reset (choose synchronous to mm_clk)
    -- MM control interface
    clk            => mm_clk,  -- : in  std_logic                     := '0';             -- control_port_clock_connection.clk
    reg_addr       => mm_mosi.address(c_tse_byte_addr_w - 1 downto 2),  -- : in  std_logic_vector(7 downto 0)  := (others => '0'); --                              .address
    reg_data_out   => mm_miso.rddata,  -- : out std_logic_vector(31 downto 0);                    --                  control_port.readdata
    reg_rd         => mm_mosi.rd,  -- : in  std_logic                     := '0';             --                              .read
    reg_data_in    => mm_mosi.wrdata,  -- : in  std_logic_vector(31 downto 0) := (others => '0'); --                              .writedata
    reg_wr         => mm_mosi.wr,  -- : in  std_logic                     := '0';             --                              .write
    reg_busy       => mm_miso.waitreq,  -- : out std_logic;                                        --                              .waitrequest
    -- Status LEDs
    led_an         => tse_led_an,  -- : out std_logic;                                        --                              .an        -- '1' = autonegation completed
    led_link       => tse_led_link,  -- : out std_logic;                                        --                              .link      -- '1' = successful link synchronisation
    led_disp_err   => OPEN,  -- : out std_logic;                                        --                              .disp_err  -- TBI character error
    led_char_err   => OPEN,  -- : out std_logic;                                        --                              .char_err  -- TBI disparity errorreceived
    led_crs        => OPEN,  -- : out std_logic;                                        --         status_led_connection.crs
    led_col        => OPEN,  -- : out std_logic;                                        --                              .col
    -- Serial 1.25 Gbps
    rx_recovclkout => OPEN,  -- : out std_logic;                                        --     serdes_control_connection.export
    ref_clk        => eth_clk,  -- : in  std_logic                     := '0';             --  pcs_ref_clk_clock_connection.clk
    txp            => eth_txp,  -- : out std_logic                                         --                              .txp_0
    rxp            => eth_rxp,  -- : in  std_logic                     := '0';             --             serial_connection.rxp_0

    tx_serial_clk      => tx_serial_clk,  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --                 tx_serial_clk.clk
    rx_cdr_refclk      => rx_cdr_refclk,  -- : in  std_logic                     := '0';             --                 rx_cdr_refclk.clk
    tx_analogreset     => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --                tx_analogreset.tx_analogreset
    tx_digitalreset    => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --               tx_digitalreset.tx_digitalreset
    rx_analogreset     => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --                rx_analogreset.rx_analogreset
    rx_digitalreset    => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --               rx_digitalreset.rx_digitalreset
    tx_cal_busy        => OPEN,  -- : out std_logic_vector(0 downto 0);                     --                   tx_cal_busy.tx_cal_busy
    rx_cal_busy        => OPEN,  -- : out std_logic_vector(0 downto 0);                     --                   rx_cal_busy.rx_cal_busy
    rx_set_locktodata  => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --             rx_set_locktodata.rx_set_locktodata
    rx_set_locktoref   => (others => '0'),  -- : in  std_logic_vector(0 downto 0)  := (others => '0'); --              rx_set_locktoref.rx_set_locktoref
    rx_is_lockedtoref  => OPEN,  -- : out std_logic_vector(0 downto 0);                     --             rx_is_lockedtoref.rx_is_lockedtoref
    rx_is_lockedtodata => open  -- : out std_logic_vector(0 downto 0)                      --            rx_is_lockedtodata.rx_is_lockedtodata
  );

  -- To be corrected
  tx_serial_clk(0) <= not tx_serial_clk(0) after serial_clk_period / 2;  -- ????
  rx_cdr_refclk    <= not rx_cdr_refclk    after cdr_clk_period / 2;  -- ????

  -- Loopback
  eth_rxp <= eth_txp;

  -- Verification
  tx_pkt_cnt <= tx_pkt_cnt + 1 when ff_tx_src_out.sop = '1' and rising_edge(dp_clk);
  rx_pkt_cnt <= rx_pkt_cnt + 1 when ff_rx_snk_in.eop = '1'  and rising_edge(dp_clk);

  p_tb_end : process
  begin
    wait until tb_end = '1';

    -- Verify that all transmitted packets have been received
    if tx_pkt_cnt = 0 then
      report "No packets were transmitted." severity ERROR;
    elsif rx_pkt_cnt = 0 then
      report "No packets were received." severity ERROR;
    elsif tx_pkt_cnt /= rx_pkt_cnt then
      report "Not all transmitted packets were received." severity ERROR;
    end if;

    -- Stop the simulation
    assert false report "Simulation finished." severity FAILURE;
    wait;
  end process;
end tb;
